# -*- coding: utf-8 -*-
"""
    emmett.asgi.protocols.helpers
    -----------------------------

    Provides HTTP protocols helpers

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import asyncio
import logging

from typing import Callable, Optional

from uvicorn.main import ServerState
from uvicorn.protocols.utils import (
    get_local_addr,
    get_remote_addr,
    is_ssl
)

from ...helpers import Config

TRACE_LOG_LEVEL = 5


class FlowControl:
    __slots__ = [
        "_is_writable_event",
        "_transport",
        "read_paused",
        "write_paused"
    ]

    def __init__(self, transport: asyncio.Transport):
        self._transport = transport
        self.read_paused = False
        self.write_paused = False
        self._is_writable_event = asyncio.Event()
        self._is_writable_event.set()

    async def drain(self):
        await self._is_writable_event.wait()

    def pause_reading(self):
        if not self.read_paused:
            self.read_paused = True
            self._transport.pause_reading()

    def resume_reading(self):
        if self.read_paused:
            self.read_paused = False
            self._transport.resume_reading()

    def pause_writing(self):
        if not self.write_paused:
            self.write_paused = True
            self._is_writable_event.clear()

    def resume_writing(self):
        if self.write_paused:
            self.write_paused = False
            self._is_writable_event.set()


class HTTPProtocol(asyncio.Protocol):
    __slots__ = [
        "access_log_enabled",
        "access_logger",
        "addr_local",
        "addr_remote",
        "app",
        "config",
        "connections",
        "default_headers",
        "flow",
        "limit_concurrency",
        "logger",
        "loop",
        "on_connection_lost",
        "root_path",
        "scheme",
        "server_state",
        "tasks",
        "timeout_keep_alive_task",
        "timeout_keep_alive",
        "transport",
        "ws_protocol_class"
    ]

    def __init__(
        self,
        config: Config,
        server_state: ServerState,
        on_connection_lost: Optional[Callable[[], None]] = None,
        _loop=None
    ):
        self.config = config
        self.app = config.loaded_app
        self.on_connection_lost = on_connection_lost
        self.loop = _loop or asyncio.get_event_loop()
        self.logger = logging.getLogger("uvicorn.error")
        self.access_logger = logging.getLogger("uvicorn.access")
        self.access_log_enabled = self.access_logger.hasHandlers()
        self.ws_protocol_class = config.ws_protocol_class
        self.root_path = config.root_path
        self.limit_concurrency = config.limit_concurrency

        # Timeouts
        self.timeout_keep_alive_task = None
        self.timeout_keep_alive = config.timeout_keep_alive

        # Shared server state
        self.server_state = server_state
        self.connections = server_state.connections
        self.tasks = server_state.tasks
        self.default_headers = server_state.default_headers

        # Per-connection state
        self.transport = None
        self.flow = None
        self.addr_local = None
        self.addr_remote = None
        self.scheme = None

    def connection_made(self, transport: asyncio.Transport):
        self.connections.add(self)

        self.transport = transport
        self.flow = FlowControl(transport)
        self.addr_local = get_local_addr(transport)
        self.addr_remote = get_remote_addr(transport)
        self.scheme = "https" if is_ssl(transport) else "http"

        if self.logger.level <= TRACE_LOG_LEVEL:
            prefix = "%s:%d - " % tuple(self.addr_remote) if self.addr_remote else ""
            self.logger.log(TRACE_LOG_LEVEL, "%sConnection made", prefix)

    def connection_lost(self, exc):
        self.connections.discard(self)

        if self.logger.level <= TRACE_LOG_LEVEL:
            prefix = "%s:%d - " % tuple(self.addr_remote) if self.addr_remote else ""
            self.logger.log(TRACE_LOG_LEVEL, "%sConnection lost", prefix)

        if self.flow is not None:
            self.flow.resume_writing()

    def eof_received(self):
        pass

    def _might_unset_keepalive(self):
        if self.timeout_keep_alive_task is not None:
            self.timeout_keep_alive_task.cancel()
            self.timeout_keep_alive_task = None

    def data_received(self, data: bytes):
        self._might_unset_keepalive()

    def on_response_complete(self):
        self.server_state.total_requests += 1
        if self.transport.is_closing():
            return

        self._might_unset_keepalive()
        self.timeout_keep_alive_task = self.loop.call_later(
            self.timeout_keep_alive, self.timeout_keep_alive_handler
        )
        self.flow.resume_reading()
        self.unblock_on_completed()

    def unblock_on_completed(self):
        pass

    def shutdown(self):
        pass

    def pause_writing(self):
        self.flow.pause_writing()

    def resume_writing(self):
        self.flow.resume_writing()

    def timeout_keep_alive_handler(self):
        pass


class ASGICycle:
    __slots__ = [
        "access_log_enabled",
        "access_logger",
        "body",
        "conn",
        "default_headers",
        "disconnected",
        "flow",
        "logger",
        "message_event",
        "more_body",
        "on_response",
        "response_completed",
        "response_started",
        "scope",
        "transport"
    ]

    def __init__(
        self,
        scope,
        conn,
        protocol: HTTPProtocol
    ):
        self.scope = scope
        self.conn = conn
        self.transport = protocol.transport
        self.flow = protocol.flow
        self.logger = protocol.logger
        self.access_logger = protocol.access_logger
        self.access_log_enabled = protocol.access_log_enabled
        self.default_headers = protocol.default_headers
        self.message_event = asyncio.Event()
        self.on_response = protocol.on_response_complete

        self.disconnected = False
        self.response_started = False
        self.response_completed = False

        self.body = b""
        self.more_body = True

    async def run_asgi(self, app):
        try:
            result = await app(self.scope, self.receive, self.send)
        except Exception as exc:
            msg = "Exception in ASGI application\n"
            self.logger.error(msg, exc_info=exc)
            if not self.response_started:
                await self.send_500_response()
            else:
                self.transport.close()
        else:
            if result is not None:
                msg = "ASGI callable should return None, but returned '%s'."
                self.logger.error(msg, result)
                self.transport.close()
            elif not self.response_started and not self.disconnected:
                msg = "ASGI callable returned without starting response."
                self.logger.error(msg)
                await self.send_500_response()
            elif not self.response_completed and not self.disconnected:
                msg = "ASGI callable returned without completing response."
                self.logger.error(msg)
                self.transport.close()
        finally:
            self.on_response = None

    async def receive(self):
        if not self.disconnected and not self.response_completed:
            self.flow.resume_reading()
            await self.message_event.wait()
            self.message_event.clear()

        if self.disconnected or self.response_completed:
            message = {"type": "http.disconnect"}
        else:
            message = {
                "type": "http.request",
                "body": self.body,
                "more_body": self.more_body,
            }
            self.body = b""

        return message

    async def send(self, message):
        raise NotImplementedError

    async def send_500_response(self):
        await self.send(
            {
                "type": "http.response.start",
                "status": 500,
                "headers": [
                    (b"content-type", b"text/plain; charset=utf-8"),
                    (b"connection", b"close")
                ]
            }
        )
        await self.send(
            {"type": "http.response.body", "body": b"Internal Server Error"}
        )


async def _service_unavailable(scope, receive, send):
    await send(
        {
            "type": "http.response.start",
            "status": 503,
            "headers": [
                (b"content-type", b"text/plain; charset=utf-8"),
                (b"connection", b"close")
            ]
        }
    )
    await send({"type": "http.response.body", "body": b"Service Unavailable"})



# -*- coding: utf-8 -*-
"""
    emmett.logger
    -------------

    Provides logging utitilites for Emmett applications.

    :copyright: 2014 Giovanni Barillari

    Based on the code of Flask (http://flask.pocoo.org)
    :copyright: (c) 2014 by Armin Ronacher.

    :license: BSD-3-Clause
"""

import logging
import os

from logging import StreamHandler, Formatter
from logging.handlers import RotatingFileHandler
from threading import Lock

from .datastructures import sdict

_logger_lock = Lock()

LOG_LEVELS = {
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'warning': logging.WARNING,
    'error': logging.ERROR,
    'critical': logging.CRITICAL
}

_def_log_config = sdict(
    production=sdict(
        max_size=5 * 1024 * 1024,
        file_no=4,
        level='warning',
        format='[%(asctime)s] %(levelname)s in %(module)s: %(message)s',
        on_app_debug=False))

_debug_log_format = (
    '> %(levelname)s in %(module)s [%(pathname)s:%(lineno)d]:\n' +
    '%(message)s'
)


def create_logger(app):
    Logger = logging.getLoggerClass()

    class DebugLogger(Logger):
        def getEffectiveLevel(x):
            if x.level == 0 and app.debug:
                return logging.DEBUG
            return Logger.getEffectiveLevel(x)

    class DebugHandler(StreamHandler):
        def emit(x, record):
            StreamHandler.emit(x, record) if app.debug else None

    class DebugRFHandler(RotatingFileHandler):
        def emit(x, record):
            RotatingFileHandler.emit(x, record) if app.debug else None

    class ProdRFHandler(RotatingFileHandler):
        def emit(x, record):
            RotatingFileHandler.emit(x, record) if not app.debug else None

    # init the console debug handler
    debug_handler = DebugHandler()
    debug_handler.setLevel(logging.DEBUG)
    debug_handler.setFormatter(Formatter(_debug_log_format))
    logger = logging.getLogger(app.logger_name)
    # just in case that was not a new logger, get rid of all the handlers
    # already attached to it.
    del logger.handlers[:]
    logger.__class__ = DebugLogger
    logger.addHandler(debug_handler)
    # load application logging config
    app_logs = app.config.logging
    if not app_logs:
        app_logs = _def_log_config
    for lname, lconf in app_logs.items():
        lfile = os.path.join(app.root_path, 'logs', lname + '.log')
        max_size = lconf.max_size or _def_log_config.production.max_size
        file_no = lconf.file_no or _def_log_config.production.file_no
        level = LOG_LEVELS.get(
            lconf.level or 'warning', LOG_LEVELS.get('warning'))
        lformat = lconf.format or _def_log_config.production.format
        on_app_debug = lconf.on_app_debug
        if on_app_debug:
            handler = DebugRFHandler(
                lfile, maxBytes=max_size, backupCount=file_no)
        else:
            handler = ProdRFHandler(
                lfile, maxBytes=max_size, backupCount=file_no)
        handler.setLevel(level)
        handler.setFormatter(Formatter(lformat))
        logger.addHandler(handler)
    return logger



from .app import App, AppModule
from .cache import Cache
from .ctx import current
from .datastructures import sdict
from .forms import Form
from .helpers import abort, stream_file
from .html import asis
from .http import redirect
from .locals import T, now, request, response, session, websocket
from .orm import Field
from .pipeline import Pipe, Injector
from .routing.urls import url



# -*- coding: utf-8 -*-
"""
    emmett.routing.response
    -----------------------

    Provides response builders for http routes.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from __future__ import annotations

from typing import Any, Dict, Union

from renoir.errors import TemplateMissingError

from ..ctx import current
from ..helpers import load_component
from ..html import asis
from ..http import HTTPResponse, HTTP, HTTPBytes
from ..wrappers.response import Response
from .rules import HTTPRoutingRule
from .urls import url

_html_content_type = 'text/html; charset=utf-8'


class MetaResponseBuilder:
    def __init__(self, route: HTTPRoutingRule):
        self.route = route

    def __call__(self, output: Any, response: Response) -> HTTPResponse:
        raise NotImplementedError


class ResponseBuilder(MetaResponseBuilder):
    http_cls = HTTP

    def __call__(self, output: Any, response: Response) -> HTTP:
        return self.http_cls(
            response.status,
            output,
            headers=response.headers,
            cookies=response.cookies
        )


class EmptyResponseBuilder(ResponseBuilder):
    http_cls = HTTPResponse

    def __call__(self, output: Any, response: Response) -> HTTPResponse:
        return self.http_cls(
            response.status,
            headers=response.headers,
            cookies=response.cookies
        )


class ResponseProcessor(ResponseBuilder):
    def process(self, output: Any, response: Response):
        raise NotImplementedError

    def __call__(self, output: Any, response: Response) -> HTTP:
        return self.http_cls(
            response.status,
            self.process(output, response),
            headers=response.headers,
            cookies=response.cookies
        )


class BytesResponseBuilder(MetaResponseBuilder):
    http_cls = HTTPBytes

    def __call__(self, output: Any, response: Response) -> HTTPBytes:
        return self.http_cls(
            response.status,
            output,
            headers=response.headers,
            cookies=response.cookies
        )


class TemplateResponseBuilder(ResponseProcessor):
    def process(
        self,
        output: Union[Dict[str, Any], None],
        response: Response
    ) -> str:
        response.headers._data['content-type'] = _html_content_type
        base_ctx = {
            'current': current,
            'url': url,
            'asis': asis,
            'load_component': load_component
        }
        output = base_ctx if output is None else {**base_ctx, **output}
        try:
            return self.route.app.templater.render(
                self.route.template, output
            )
        except TemplateMissingError as exc:
            raise HTTP(
                404,
                body="{}\n".format(exc.message),
                cookies=response.cookies
            )


class AutoResponseBuilder(ResponseProcessor):
    def process(self, output: Any, response: Response) -> str:
        is_template = False
        if isinstance(output, dict):
            is_template = True
            output = {
                **{
                    'current': current,
                    'url': url,
                    'asis': asis,
                    'load_component': load_component
                },
                **output
            }
        elif output is None:
            output = {
                'current': current,
                'url': url,
                'asis': asis,
                'load_component': load_component
            }
        if is_template:
            response.headers._data['content-type'] = _html_content_type
            try:
                return self.route.app.templater.render(
                    self.route.template, output
                )
            except TemplateMissingError as exc:
                raise HTTP(
                    404,
                    body="{}\n".format(exc.message),
                    cookies=response.cookies
                )
        elif isinstance(output, str):
            return output
        return str(output)



# -*- coding: utf-8 -*-
"""
    tests.pipeline
    --------------

    Test Emmett pipeline
"""

import asyncio
import pytest

from contextlib import contextmanager

from helpers import current_ctx as _current_ctx, ws_ctx as _ws_ctx
from emmett import App, request, websocket, abort
from emmett.ctx import current
from emmett.http import HTTP
from emmett.pipeline import Pipe, Injector
from emmett.parsers import Parsers
from emmett.serializers import Serializers, _json_type

json_load = Parsers.get_for('json')
json_dump = Serializers.get_for('json')


class PipeException(Exception):
    def __init__(self, pipe):
        self.pipe = pipe


class FlowStorePipe(Pipe):
    @property
    def linear_storage(self):
        return current._pipeline_linear_storage

    @property
    def parallel_storage(self):
        return current._pipeline_parallel_storage

    def store_linear(self, status):
        self.linear_storage.append(self.__class__.__name__ + "." + status)

    def store_parallel(self, status):
        self.parallel_storage.append(self.__class__.__name__ + "." + status)

    async def on_pipe_success(self):
        self.store_linear('success')

    async def on_pipe_failure(self):
        self.store_linear('failure')


class FlowStorePipeCommon(FlowStorePipe):
    async def open(self):
        self.store_parallel('open')

    async def close(self):
        self.store_parallel('close')

    async def pipe(self, next_pipe, **kwargs):
        self.store_linear('pipe')
        return await next_pipe(**kwargs)


class FlowStorePipeSplit(FlowStorePipe):
    async def open_request(self):
        self.store_parallel('open_request')

    async def open_ws(self):
        self.store_parallel('open_ws')

    async def close_request(self):
        self.store_parallel('close_request')

    async def close_ws(self):
        self.store_parallel('close_ws')

    async def pipe_request(self, next_pipe, **kwargs):
        self.store_linear('pipe_request')
        return await next_pipe(**kwargs)

    async def pipe_ws(self, next_pipe, **kwargs):
        self.store_linear('pipe_ws')
        return await next_pipe(**kwargs)


class Pipe1(FlowStorePipeCommon):
    pass


class Pipe2(FlowStorePipeSplit):
    async def pipe_request(self, next_pipe, **kwargs):
        self.store_linear('pipe_request')
        if request.query_params.skip:
            return "block"
        return await next_pipe(**kwargs)

    async def pipe_ws(self, next_pipe, **kwargs):
        self.store_linear('pipe_ws')
        if websocket.query_params.skip:
            return
        await next_pipe(**kwargs)


class Pipe3(FlowStorePipeCommon):
    async def open(self):
        await asyncio.sleep(0.05)
        await super().open()


class Pipe4(FlowStorePipeCommon):
    async def close(self):
        await asyncio.sleep(0.05)
        await super().close()


class Pipe5(FlowStorePipeCommon):
    pass


class Pipe6(FlowStorePipeCommon):
    pass


class ExcPipeOpen(FlowStorePipeCommon):
    async def open(self):
        raise PipeException(self)


class ExcPipeClose(FlowStorePipeCommon):
    async def close(self):
        raise PipeException(self)


class PipeSR1(FlowStorePipeSplit):
    def on_receive(self, data):
        data = json_load(data)
        return dict(pipe1r='receive_inject', **data)

    def on_send(self, data):
        return json_dump(dict(pipe1s='send_inject', **data))


class PipeSR2(FlowStorePipeSplit):
    def on_receive(self, data):
        return dict(pipe2r='receive_inject', **data)

    def on_send(self, data):
        return dict(pipe2s='send_inject', **data)


class CTXInjector(Injector):
    async def pipe_request(self, next_pipe, **kwargs):
        rv = await super().pipe_request(next_pipe, **kwargs)
        current._pipeline_generic_storage.append(rv)
        return rv


class GlobalInjector(CTXInjector):
    foo = "bar"

    def __init__(self):
        self.bar = "baz"
        super().__init__()

    @staticmethod
    def staticm(val):
        return val

    def boundm(self, val):
        return self.foo, val

    @property
    def prop(self):
        return self.bar


class ScopedInjector(CTXInjector):
    namespace = "scoped"
    foo = "bar"

    def __init__(self):
        self.bar = "baz"
        super().__init__()

    @staticmethod
    def staticm(val):
        return val

    def boundm(self, val):
        return self.foo, val

    @property
    def prop(self):
        return self.bar


class CTXWrapper:
    def __init__(self, ctx, router):
        self.ctx = ctx
        self.router = router

    def dispatch(self):
        return self.router.dispatch()


class ReqCTXWrapper(CTXWrapper):
    def dispatch(self):
        return self.router.dispatch(self.ctx.request, self.ctx.response)


class WSCTXWrapper(CTXWrapper):
    def dispatch(self):
        return self.router.dispatch(self.ctx.websocket)


@contextmanager
def request_ctx(app, path):
    with _current_ctx(path) as ctx:
        ctx._pipeline_generic_storage = []
        ctx._pipeline_linear_storage = []
        ctx._pipeline_parallel_storage = []
        yield ReqCTXWrapper(ctx, app._router_http)


@contextmanager
def ws_ctx(app, path):
    with _ws_ctx(path) as ctx:
        ctx._pipeline_generic_storage = []
        ctx._pipeline_linear_storage = []
        ctx._pipeline_parallel_storage = []
        yield WSCTXWrapper(ctx, app._router_ws)


def linear_flows_are_equal(flow, ctx):
    try:
        for index, value in enumerate(flow):
            if ctx._pipeline_linear_storage[index] != value:
                return False
    except Exception:
        return False
    return True


def parallel_flows_are_equal(flow, ctx):
    return set(flow) == set(ctx._pipeline_parallel_storage)


@pytest.fixture(scope='module')
def app():
    app = App(__name__)
    app.pipeline = [Pipe1(), Pipe2(), Pipe3()]

    @app.route()
    def ok():
        return "ok"

    @app.route()
    def http_error():
        abort(422)

    @app.route()
    def error():
        raise Exception

    @app.route(pipeline=[ExcPipeOpen(), Pipe4()])
    def open_error():
        return ''

    @app.route(pipeline=[ExcPipeClose(), Pipe4()])
    def close_error():
        return ''

    @app.route(pipeline=[Pipe4()])
    def pipe4():
        return "4"

    @app.websocket()
    async def ws_ok():
        await websocket.send('ok')

    @app.websocket()
    def ws_error():
        raise Exception

    @app.websocket(pipeline=[ExcPipeOpen(), Pipe4()])
    def ws_open_error():
        return

    @app.websocket(pipeline=[ExcPipeClose(), Pipe4()])
    def ws_close_error():
        return

    @app.websocket(pipeline=[Pipe4()])
    def ws_pipe4():
        return

    @app.websocket(pipeline=[PipeSR1(), PipeSR2()])
    async def ws_inject():
        data = await websocket.receive()
        current._receive_storage.append(data)
        await websocket.send(data)

    mod = app.module(__name__, 'mod', url_prefix='mod')
    mod.pipeline = [Pipe5()]

    @mod.route()
    def pipe5():
        return "5"

    @mod.route(pipeline=[Pipe6()])
    def pipe6():
        return "6"

    @mod.websocket()
    def ws_pipe5():
        return

    @mod.websocket(pipeline=[Pipe6()])
    def ws_pipe6():
        return

    inj = app.module(__name__, 'inj', url_prefix='inj')
    inj.pipeline = [GlobalInjector(), ScopedInjector()]

    @inj.route(template='test.html')
    def injpipe():
        return {'posts': []}

    return app


@pytest.mark.asyncio
async def test_ok_flow(app):
    with request_ctx(app, '/ok') as ctx:
        parallel_flow = [
            'Pipe1.open', 'Pipe2.open_request', 'Pipe3.open',
            'Pipe3.close', 'Pipe2.close_request', 'Pipe1.close']
        linear_flow = [
            'Pipe1.pipe', 'Pipe2.pipe_request', 'Pipe3.pipe',
            'Pipe3.success', 'Pipe2.success', 'Pipe1.success']
        await ctx.dispatch()
        assert linear_flows_are_equal(linear_flow, ctx.ctx)
        assert parallel_flows_are_equal(parallel_flow, ctx.ctx)

    with ws_ctx(app, '/ws_ok') as ctx:
        parallel_flow = [
            'Pipe1.open', 'Pipe2.open_ws', 'Pipe3.open',
            'Pipe3.close', 'Pipe2.close_ws', 'Pipe1.close']
        linear_flow = [
            'Pipe1.pipe', 'Pipe2.pipe_ws', 'Pipe3.pipe',
            'Pipe3.success', 'Pipe2.success', 'Pipe1.success']
        await ctx.dispatch()
        assert linear_flows_are_equal(linear_flow, ctx.ctx)
        assert parallel_flows_are_equal(parallel_flow, ctx.ctx)


@pytest.mark.asyncio
async def test_httperror_flow(app):
    with request_ctx(app, '/http_error') as ctx:
        parallel_flow = [
            'Pipe1.open', 'Pipe2.open_request', 'Pipe3.open',
            'Pipe3.close', 'Pipe2.close_request', 'Pipe1.close']
        linear_flow = [
            'Pipe1.pipe', 'Pipe2.pipe_request', 'Pipe3.pipe',
            'Pipe3.success', 'Pipe2.success', 'Pipe1.success']
        try:
            await ctx.dispatch()
        except HTTP:
            pass
        assert linear_flows_are_equal(linear_flow, ctx.ctx)
        assert parallel_flows_are_equal(parallel_flow, ctx.ctx)


@pytest.mark.asyncio
async def test_error_flow(app):
    with request_ctx(app, '/error') as ctx:
        parallel_flow = [
            'Pipe1.open', 'Pipe2.open_request', 'Pipe3.open',
            'Pipe3.close', 'Pipe2.close_request', 'Pipe1.close']
        linear_flow = [
            'Pipe1.pipe', 'Pipe2.pipe_request', 'Pipe3.pipe',
            'Pipe3.failure', 'Pipe2.failure', 'Pipe1.failure']
        try:
            await ctx.dispatch()
        except Exception:
            pass
        assert linear_flows_are_equal(linear_flow, ctx.ctx)
        assert parallel_flows_are_equal(parallel_flow, ctx.ctx)

    with ws_ctx(app, '/ws_error') as ctx:
        parallel_flow = [
            'Pipe1.open', 'Pipe2.open_ws', 'Pipe3.open',
            'Pipe3.close', 'Pipe2.close_ws', 'Pipe1.close']
        linear_flow = [
            'Pipe1.pipe', 'Pipe2.pipe_ws', 'Pipe3.pipe',
            'Pipe3.failure', 'Pipe2.failure', 'Pipe1.failure']
        try:
            await ctx.dispatch()
        except Exception:
            pass
        assert linear_flows_are_equal(linear_flow, ctx.ctx)
        assert parallel_flows_are_equal(parallel_flow, ctx.ctx)


@pytest.mark.asyncio
async def test_open_error(app):
    with request_ctx(app, '/open_error') as ctx:
        parallel_flow = [
            'Pipe1.open', 'Pipe2.open_request', 'Pipe3.open', 'Pipe4.open']
        linear_flow = []
        try:
            await ctx.dispatch()
        except PipeException as e:
            assert isinstance(e.pipe, ExcPipeOpen)
        assert linear_flows_are_equal(linear_flow, ctx.ctx)
        assert parallel_flows_are_equal(parallel_flow, ctx.ctx)

    with ws_ctx(app, '/ws_open_error') as ctx:
        parallel_flow = [
            'Pipe1.open', 'Pipe2.open_ws', 'Pipe3.open', 'Pipe4.open']
        linear_flow = []
        try:
            await ctx.dispatch()
        except PipeException as e:
            assert isinstance(e.pipe, ExcPipeOpen)
        assert linear_flows_are_equal(linear_flow, ctx.ctx)
        assert parallel_flows_are_equal(parallel_flow, ctx.ctx)


@pytest.mark.asyncio
async def test_close_error(app):
    with request_ctx(app, '/close_error') as ctx:
        parallel_flow = [
            'Pipe1.open', 'Pipe2.open_request', 'Pipe3.open',
            'ExcPipeClose.open', 'Pipe4.open',
            'Pipe4.close', 'Pipe3.close', 'Pipe2.close_request', 'Pipe1.close']
        linear_flow = [
            'Pipe1.pipe', 'Pipe2.pipe_request', 'Pipe3.pipe',
            'ExcPipeClose.pipe', 'Pipe4.pipe',
            'Pipe4.success', 'ExcPipeClose.success', 'Pipe3.success',
            'Pipe2.success', 'Pipe1.success']
        try:
            await ctx.dispatch()
        except PipeException as e:
            assert isinstance(e.pipe, ExcPipeClose)
        assert linear_flows_are_equal(linear_flow, ctx.ctx)
        assert parallel_flows_are_equal(parallel_flow, ctx.ctx)

    with ws_ctx(app, '/ws_close_error') as ctx:
        parallel_flow = [
            'Pipe1.open', 'Pipe2.open_ws', 'Pipe3.open',
            'ExcPipeClose.open', 'Pipe4.open',
            'Pipe4.close', 'Pipe3.close', 'Pipe2.close_ws', 'Pipe1.close']
        linear_flow = [
            'Pipe1.pipe', 'Pipe2.pipe_ws', 'Pipe3.pipe',
            'ExcPipeClose.pipe', 'Pipe4.pipe',
            'Pipe4.success', 'ExcPipeClose.success', 'Pipe3.success',
            'Pipe2.success', 'Pipe1.success']
        try:
            await ctx.dispatch()
        except PipeException as e:
            assert isinstance(e.pipe, ExcPipeClose)
        assert linear_flows_are_equal(linear_flow, ctx.ctx)
        assert parallel_flows_are_equal(parallel_flow, ctx.ctx)


@pytest.mark.asyncio
async def test_flow_interrupt(app):
    with request_ctx(app, '/ok?skip=yes') as ctx:
        parallel_flow = [
            'Pipe1.open', 'Pipe2.open_request', 'Pipe3.open',
            'Pipe3.close', 'Pipe2.close_request', 'Pipe1.close']
        linear_flow = [
            'Pipe1.pipe', 'Pipe2.pipe_request',
            'Pipe2.success', 'Pipe1.success']
        await ctx.dispatch()
        assert linear_flows_are_equal(linear_flow, ctx.ctx)
        assert parallel_flows_are_equal(parallel_flow, ctx.ctx)

    with ws_ctx(app, '/ws_ok?skip=yes') as ctx:
        parallel_flow = [
            'Pipe1.open', 'Pipe2.open_ws', 'Pipe3.open',
            'Pipe3.close', 'Pipe2.close_ws', 'Pipe1.close']
        linear_flow = [
            'Pipe1.pipe', 'Pipe2.pipe_ws',
            'Pipe2.success', 'Pipe1.success']
        await ctx.dispatch()
        assert linear_flows_are_equal(linear_flow, ctx.ctx)
        assert parallel_flows_are_equal(parallel_flow, ctx.ctx)


@pytest.mark.asyncio
async def test_pipeline_composition(app):
    with request_ctx(app, '/pipe4') as ctx:
        parallel_flow = [
            'Pipe1.open', 'Pipe2.open_request', 'Pipe3.open', 'Pipe4.open',
            'Pipe4.close', 'Pipe3.close', 'Pipe2.close_request', 'Pipe1.close']
        linear_flow = [
            'Pipe1.pipe', 'Pipe2.pipe_request', 'Pipe3.pipe', 'Pipe4.pipe',
            'Pipe4.success', 'Pipe3.success', 'Pipe2.success', 'Pipe1.success']
        await ctx.dispatch()
        assert linear_flows_are_equal(linear_flow, ctx.ctx)
        assert parallel_flows_are_equal(parallel_flow, ctx.ctx)

    with ws_ctx(app, '/ws_pipe4') as ctx:
        parallel_flow = [
            'Pipe1.open', 'Pipe2.open_ws', 'Pipe3.open', 'Pipe4.open',
            'Pipe4.close', 'Pipe3.close', 'Pipe2.close_ws', 'Pipe1.close']
        linear_flow = [
            'Pipe1.pipe', 'Pipe2.pipe_ws', 'Pipe3.pipe', 'Pipe4.pipe',
            'Pipe4.success', 'Pipe3.success', 'Pipe2.success', 'Pipe1.success']
        await ctx.dispatch()
        assert linear_flows_are_equal(linear_flow, ctx.ctx)
        assert parallel_flows_are_equal(parallel_flow, ctx.ctx)


@pytest.mark.asyncio
async def test_module_pipeline(app):
    with request_ctx(app, '/mod/pipe5') as ctx:
        parallel_flow = [
            'Pipe1.open', 'Pipe2.open_request', 'Pipe3.open', 'Pipe5.open',
            'Pipe5.close', 'Pipe3.close', 'Pipe2.close_request', 'Pipe1.close']
        linear_flow = [
            'Pipe1.pipe', 'Pipe2.pipe_request', 'Pipe3.pipe', 'Pipe5.pipe',
            'Pipe5.success', 'Pipe3.success', 'Pipe2.success', 'Pipe1.success']
        await ctx.dispatch()
        assert linear_flows_are_equal(linear_flow, ctx.ctx)
        assert parallel_flows_are_equal(parallel_flow, ctx.ctx)

    with ws_ctx(app, '/mod/ws_pipe5') as ctx:
        parallel_flow = [
            'Pipe1.open', 'Pipe2.open_ws', 'Pipe3.open', 'Pipe5.open',
            'Pipe5.close', 'Pipe3.close', 'Pipe2.close_ws', 'Pipe1.close']
        linear_flow = [
            'Pipe1.pipe', 'Pipe2.pipe_ws', 'Pipe3.pipe', 'Pipe5.pipe',
            'Pipe5.success', 'Pipe3.success', 'Pipe2.success', 'Pipe1.success']
        await ctx.dispatch()
        assert linear_flows_are_equal(linear_flow, ctx.ctx)
        assert parallel_flows_are_equal(parallel_flow, ctx.ctx)


@pytest.mark.asyncio
async def test_module_pipeline_composition(app):
    with request_ctx(app, '/mod/pipe6') as ctx:
        parallel_flow = [
            'Pipe1.open', 'Pipe2.open_request', 'Pipe3.open', 'Pipe5.open',
            'Pipe6.open',
            'Pipe6.close', 'Pipe5.close', 'Pipe3.close', 'Pipe2.close_request',
            'Pipe1.close']
        linear_flow = [
            'Pipe1.pipe', 'Pipe2.pipe_request', 'Pipe3.pipe', 'Pipe5.pipe',
            'Pipe6.pipe',
            'Pipe6.success', 'Pipe5.success', 'Pipe3.success', 'Pipe2.success',
            'Pipe1.success']
        await ctx.dispatch()
        assert linear_flows_are_equal(linear_flow, ctx.ctx)
        assert parallel_flows_are_equal(parallel_flow, ctx.ctx)

    with ws_ctx(app, '/mod/ws_pipe6') as ctx:
        parallel_flow = [
            'Pipe1.open', 'Pipe2.open_ws', 'Pipe3.open', 'Pipe5.open',
            'Pipe6.open',
            'Pipe6.close', 'Pipe5.close', 'Pipe3.close', 'Pipe2.close_ws',
            'Pipe1.close']
        linear_flow = [
            'Pipe1.pipe', 'Pipe2.pipe_ws', 'Pipe3.pipe', 'Pipe5.pipe',
            'Pipe6.pipe',
            'Pipe6.success', 'Pipe5.success', 'Pipe3.success', 'Pipe2.success',
            'Pipe1.success']
        await ctx.dispatch()
        assert linear_flows_are_equal(linear_flow, ctx.ctx)
        assert parallel_flows_are_equal(parallel_flow, ctx.ctx)


@pytest.mark.asyncio
async def test_receive_send_flow(app):
    send_storage_key = {
        "str": "text",
        "bytes": "bytes"
    }[_json_type]
    with ws_ctx(app, '/ws_inject') as ctx:
        parallel_flow = [
            'Pipe1.open', 'Pipe2.open_ws', 'Pipe3.open',
            'PipeSR1.open_ws', 'PipeSR2.open_ws',
            'PipeSR2.close_ws', 'PipeSR1.close_ws',
            'Pipe3.close', 'Pipe2.close_ws', 'Pipe1.close']
        linear_flow = [
            'Pipe1.pipe', 'Pipe2.pipe_ws', 'Pipe3.pipe',
            'PipeSR1.pipe_ws', 'PipeSR2.pipe_ws',
            'PipeSR2.success', 'PipeSR1.success',
            'Pipe3.success', 'Pipe2.success', 'Pipe1.success']
        await ctx.dispatch()
        assert linear_flows_are_equal(linear_flow, ctx.ctx)
        assert parallel_flows_are_equal(parallel_flow, ctx.ctx)

        assert ctx.ctx._receive_storage[-1] == {
            'foo': 'bar',
            'pipe1r': 'receive_inject', 'pipe2r': 'receive_inject'
        }
        assert json_load(ctx.ctx._send_storage[-1][send_storage_key]) == {
            'foo': 'bar',
            'pipe1r': 'receive_inject', 'pipe2r': 'receive_inject',
            'pipe1s': 'send_inject', 'pipe2s': 'send_inject'
        }


@pytest.mark.asyncio
async def test_injectors(app):
    with request_ctx(app, '/inj/injpipe') as ctx:
        current.app = app

        await ctx.dispatch()
        env = ctx.ctx._pipeline_generic_storage[0]

        assert env['posts'] == []
        assert env['foo'] == "bar"
        assert env['bar'] == "baz"
        assert env['staticm']("test") == "test"
        assert env['boundm']("test") == ("bar", "test")
        assert env['prop'] == "baz"

        env = env['scoped']
        assert env.foo == "bar"
        assert env.bar == "baz"
        assert env.staticm("test") == "test"
        assert env.boundm("test") == ("bar", "test")
        assert env.prop == "baz"



# -*- coding: utf-8 -*-
"""
    emmett.orm.migrations.helpers
    -----------------------------

    Provides helpers for migrations.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from collections.abc import Iterable
from contextlib import contextmanager
from uuid import uuid4

from ...datastructures import _unique_list


DEFAULT_VALUE = lambda: None


def make_migration_id():
    return uuid4().hex[-12:]


class WrappedOperation(object):
    def __init__(self, op_class, name, engine):
        self.op_class = op_class
        self.name = name
        self.engine = engine

    def __call__(self, *args, **kwargs):
        op = getattr(self.op_class, self.name)(*args, **kwargs)
        op._env_load_(self.engine)
        return op.run()


class Dispatcher(object):
    def __init__(self):
        self._registry = {}

    def dispatch_for(self, target):
        def wrap(fn):
            self._registry[target] = fn
            return fn
        return wrap

    def dispatch(self, obj):
        targets = type(obj).__mro__
        for target in targets:
            if target in self._registry:
                return self._registry[target]
        else:
            raise ValueError("no dispatch function for object: %s" % obj)


class DryRunAdapter:
    def __init__(self, adapter, logger):
        self.adapter = adapter
        self.__dlogger = logger

    def __getattr__(self, name):
        return getattr(self.adapter, name)

    def execute(self, sql):
        self.__dlogger(sql)


class DryRunDatabase:
    def __init__(self, db, logger):
        self.db = db
        self._adapter = DryRunAdapter(db._adapter, logger)

    def __getattr__(self, name):
        return getattr(self.db, name)

    def __getitem__(self, key):
        return self.db[key]

    @contextmanager
    def connection(self, *args, **kwargs):
        yield None


def to_tuple(x, default=None):
    if x is None:
        return default
    elif isinstance(x, str):
        return (x, )
    elif isinstance(x, Iterable):
        return tuple(x)
    else:
        return (x, )


def tuple_or_value(val):
    if not val:
        return None
    elif len(val) == 1:
        return val[0]
    else:
        return val


def tuple_rev_as_scalar(rev):
    if not rev:
        return None
    elif len(rev) == 1:
        return rev[0]
    else:
        return rev


def dedupe_tuple(tup):
    return tuple(_unique_list(tup))


def format_with_comma(value):
    if value is None:
        return ""
    elif isinstance(value, str):
        return value
    elif isinstance(value, Iterable):
        return ", ".join(value)
    else:
        raise ValueError("Don't know how to comma-format %r" % value)


def _feasible_as_dbms_default(val):
    if callable(val):
        return False
    if val is None:
        return True
    if isinstance(val, int):
        return True
    if isinstance(val, str):
        return True
    if isinstance(val, (bool, float)):
        return True
    return False



# -*- coding: utf-8 -*-
"""
    emmett.validators.process
    -------------------------

    Validators that transform values.

    Ported from the original validators of web2py (http://www.web2py.com)

    :copyright: (c) by Massimo Di Pierro <mdipierro@cs.depaul.edu>
    :license: LGPLv3 (http://www.gnu.org/licenses/lgpl.html)
"""

import re
import unicodedata

# TODO: check unicode conversions
from .._shortcuts import to_unicode
from .basic import Validator
from .helpers import translate, LazyCrypt


class Cleanup(Validator):
    rule = re.compile(r"[^\x09\x0a\x0d\x20-\x7e]")

    def __init__(self, regex=None, message=None):
        super().__init__(message=message)
        self.regex = self.rule if regex is None else re.compile(regex)

    def __call__(self, value):
        v = self.regex.sub('', (to_unicode(value) or '').strip())
        return v, None


class Lower(Validator):
    def __call__(self, value):
        if value is None:
            return (value, None)
        return (to_unicode(value).lower(), None)


class Upper(Validator):
    def __call__(self, value):
        if value is None:
            return (value, None)
        return (to_unicode(value).upper(), None)


class Urlify(Validator):
    message = "Not convertible to url"

    def __init__(self, maxlen=80, check=False, keep_underscores=False, message=None):
        super().__init__(message=message)
        self.maxlen = maxlen
        self.check = check
        self.message = message
        self.keep_underscores = keep_underscores

    def __call__(self, value):
        if self.check and value != self._urlify(value):
            return value, translate(self.message)
        return self._urlify(value), None

    def _urlify(self, s):
        """
        Converts incoming string to a simplified ASCII subset.
        if (keep_underscores): underscores are retained in the string
        else: underscores are translated to hyphens (default)
        """
        s = to_unicode(s)
        # to lowercase
        s = s.lower()
        # replace special characters
        s = unicodedata.normalize("NFKD", s)
        # encode as ASCII
        s = s.encode("ascii", "ignore").decode("ascii")
        # strip html entities
        s = re.sub(r"&\w+?;", "", s)
        if self.keep_underscores:
            # whitespace to hypens
            s = re.sub(r"\s+", "-", s)
            # strip all but alphanumeric/underscore/hyphen
            s = re.sub(r"[^\w\-]", "", s)
        else:
            # whitespace & underscores to hyphens
            s = re.sub(r"[\s_]+", "-", s)
            # strip all but alphanumeric/hyphen
            s = re.sub(r"[^a-z0-9\-]", "", s)
        # collapse strings of hyphens
        s = re.sub(r"[-_][-_]+", "-", s)
        # remove leading and trailing hyphens
        s = s.strip(r"-")
        # enforce maximum length
        return s[:self.maxlen]


class Crypt(Validator):
    """
    encodes the value on validation with a digest.

    If no arguments are provided Crypt uses the MD5 algorithm.
    If the key argument is provided the HMAC+MD5 algorithm is used.
    If the digest_alg is specified this is used to replace the
    MD5 with, for example, SHA512. The digest_alg can be
    the name of a hashlib algorithm as a string or the algorithm itself.

    min_length is the minimal password length (default 4)
    error_message is the message if password is too short

    Notice that an empty password is accepted but invalid. It will not allow
    login back. Stores junk as hashed password.

    Specify an algorithm or by default we will use sha512.

    Typical available algorithms:
      md5, sha1, sha224, sha256, sha384, sha512

    If salt, it hashes a password with a salt.
    If salt is True, this method will automatically generate one.
    Either case it returns an encrypted password string with format:

      <algorithm>$<salt>$<hash>

    Important: hashed password is returned as a LazyCrypt object and computed
    only if needed. The LasyCrypt object also knows how to compare itself with
    an existing salted password
    """

    def __init__(
        self, key=None, algorithm='pbkdf2(1000,20,sha512)', salt=True, message=None
    ):
        super().__init__(message=message)
        self.key = key
        self.digest_alg = algorithm
        self.salt = salt

    def __call__(self, value):
        return LazyCrypt(self, value), None



# -*- coding: utf-8 -*-
"""
    emmett.ctx
    ----------

    Provides the current object.
    Used by application to deal with request related objects.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import contextvars

from datetime import datetime
from typing import Any

import pendulum

from .utils import cachedprop

_ctxv = contextvars.ContextVar("_emt_ctxv")


class Context:
    __slots__ = ["app", "__dict__"]

    def __init__(self):
        self.language = None

    @property
    def now(self):
        return pendulum.instance(datetime.utcnow())


class RequestContext(Context):
    __slots__ = ["request", "response", "session"]

    def __init__(
        self,
        app,
        scope,
        receive,
        send,
        wrapper_request,
        wrapper_response
    ):
        self.app = app
        self.request = wrapper_request(
            scope,
            receive,
            send,
            app.config.request_max_content_length,
            app.config.request_body_timeout
        )
        self.response = wrapper_response()
        self.session = None

    @property
    def now(self):
        return self.request.now

    @cachedprop
    def language(self):
        return self.request.accept_language.best_match(
            list(self.app.translator._langmap)
        )


class WSContext(Context):
    __slots__ = ["websocket", "session"]

    def __init__(self, app, scope, receive, send, wrapper_websocket):
        self.app = app
        self.websocket = wrapper_websocket(
            scope,
            receive,
            send
        )
        self.session = None

    @cachedprop
    def language(self):
        return self.websocket.accept_language.best_match(
            list(self.app.translator._langmap)
        )


class Current:
    __slots__ = []

    ctx = property(_ctxv.get)

    def __init__(self):
        _ctxv.set(Context())

    def _init_(self, ctx):
        return _ctxv.set(ctx)

    def _close_(self, token):
        _ctxv.reset(token)

    def __getattr__(self, name: str) -> Any:
        return getattr(self.ctx, name)

    def __setattr__(self, name: str, value: Any):
        setattr(self.ctx, name, value)

    def __delattr__(self, name: str):
        delattr(self.ctx, name)

    def __getitem__(self, name: str) -> Any:
        try:
            return getattr(self.ctx, name)
        except AttributeError as e:
            raise KeyError from e

    def __setitem__(self, name: str, value: Any):
        setattr(self.ctx, name, value)

    def __delitem__(self, name: str):
        delattr(self.ctx, name)

    def __contains__(self, name: str) -> bool:
        return hasattr(self.ctx, name)

    def get(self, name: str, default: Any = None) -> Any:
        return getattr(self.ctx, name, default)

    @property
    def T(self):
        return self.ctx.app.translator


current = Current()



# -*- coding: utf-8 -*-
"""
    tests.validators
    ----------------

    Test Emmett validators over pyDAL.
"""

import pytest

from datetime import datetime, timedelta
from emmett import App, sdict
from emmett.orm import Database, Model, Field, has_many, belongs_to
from emmett.validators import (
    isEmptyOr, hasLength, isInt, isFloat, isDate, isTime, isDatetime, isJSON,
    isntEmpty, inSet, inDB, isEmail, isUrl, isIP, isImage, inRange, Equals,
    Lower, Upper, Cleanup, Urlify, Crypt, notInDB, Allow, Not, Matches, Any,
    isList)


class A(Model):
    tablename = "a"

    name = Field()
    val = Field.int()
    fval = Field.float()
    text = Field.text()
    password = Field.password()
    d = Field.date()
    t = Field.time()
    dt = Field.datetime()
    json = Field.json()


class AA(Model):
    tablename = "aa"

    a = Field(auto_validation=False)


class AAA(Model):
    tablename = "aaa"
    auto_validation = False

    a = Field()


class B(Model):
    tablename = "b"

    a = Field()
    b = Field(validation={'len': {'gte': 5}})

    validation = {
        'a': {'len': {'gte': 5}}
    }


class Consist(Model):
    email = Field()
    url = Field()
    ip = Field()
    image = Field.upload()
    emails = Field.string_list()
    emailsplit = Field.string_list()

    validation = {
        'email': {'is': 'email'},
        'url': {'is': 'url'},
        'ip': {'is': 'ip'},
        'image': {'is': 'image'},
        'emails': {'is': 'list:email'},
        'emailsplit': {'is': {'list:email': {'splitter': ',;'}}}
    }


class Len(Model):
    a = Field()
    b = Field()
    c = Field()
    d = Field()

    validation = {
        'a': {'len': 5},
        'b': {'len': {'gt': 4, 'lt': 13}},
        'c': {'len': {'gte': 5, 'lte': 12}},
        'd': {'len': {'range': (5, 13)}}
    }


class Inside(Model):
    a = Field()
    b = Field.int()

    validation = {
        'a': {'in': ['a', 'b']},
        'b': {'in': {'range': (1, 5)}}
    }


class Num(Model):
    a = Field.int()
    b = Field.int()
    c = Field.int()

    validation = {
        'a': {'gt': 0},
        'b': {'lt': 5},
        'c': {'gt': 0, 'lte': 4}
    }


class Proc(Model):
    a = Field()
    b = Field()
    c = Field()
    d = Field()
    e = Field.password()
    f = Field.password()

    validation = {
        'a': {'lower': True},
        'b': {'upper': True},
        'c': {'clean': True},
        'd': {'urlify': True},
        'e': {'len': {'range': (6, 25)}, 'crypt': True},
        'f': {'len': {'gt': 5, 'lt': 25}, 'crypt': 'md5'}
    }


class Eq(Model):
    a = Field()
    b = Field.int()
    c = Field.float()

    validation = {
        'a': {'equals': 'asd'},
        'b': {'equals': 2},
        'c': {'not': {'equals': 2.4}}
    }


class Match(Model):
    a = Field()
    b = Field()

    validation = {
        'a': {'match': 'ab'},
        'b': {'match': {'expression': 'ab', 'strict': True}}
    }


class Anyone(Model):
    a = Field()

    validation = {
        'a': {'any': {'is': 'email', 'in': ['foo', 'bar']}}
    }


class Person(Model):
    has_many('things')

    name = Field(validation={'empty': False})
    surname = Field(validation={'presence': True})


class Thing(Model):
    belongs_to('person')

    name = Field()
    color = Field()
    uid = Field(unique=True)

    validation = {
        'name': {'presence': True},
        'color': {'in': ['blue', 'red']},
        'uid': {'empty': False}
    }


class Allowed(Model):
    a = Field(validation={'in': ['a', 'b'], 'allow': None})
    b = Field(validation={'in': ['a', 'b'], 'allow': 'empty'})
    c = Field(validation={'in': ['a', 'b'], 'allow': 'blank'})


class Mixed(Model):
    belongs_to('person')

    date = Field.date()
    type = Field()
    inside = Field()
    number = Field.int()
    dont = Field()
    yep = Field()
    psw = Field.password()

    validation = {
        'date': {'format': '%d/%m/%Y', 'gt': lambda: datetime.utcnow().date()},
        'type': {'in': ['a', 'b'], 'allow': None},
        'inside': {'in': ['asd', 'lol']},
        'number': {'allow': 'blank'},
        'dont': {'empty': True},
        'yep': {'presence': True},
        'psw': {'len': {'range': (6, 25)}, 'crypt': True}
    }


@pytest.fixture(scope='module')
def db():
    app = App(__name__)
    db = Database(
        app, config=sdict(
            uri='sqlite://validators.db', auto_connect=True,
            auto_migrate=True))
    db.define_models([
        A, AA, AAA, B, Consist, Len, Inside, Num, Eq, Match, Anyone, Proc,
        Person, Thing, Allowed, Mixed
    ])
    return db


def test_defaults(db):
    #: string, text, password
    assert isinstance(db.a.name.requires[0], hasLength)
    assert db.a.name.requires[0].minsize == 0
    assert db.a.name.requires[0].maxsize == db.a.name.length
    assert isinstance(db.a.text.requires[0], hasLength)
    assert isinstance(db.a.password.requires[0], hasLength)
    #: numbers
    assert isinstance(db.a.val.requires[0], isInt)
    assert isinstance(db.a.fval.requires[0], isFloat)
    #: date, time, datetime
    assert isinstance(db.a.d.requires[0], isDate)
    assert isinstance(db.a.t.requires[0], isTime)
    assert isinstance(db.a.dt.requires[0], isDatetime)
    #: json
    assert isinstance(db.a.json.requires[0], isJSON)


def test_defaults_disable(db):
    assert len(db.aa.a.requires) == 0
    assert len(db.aaa.a.requires) == 0


def test_requires_vs_validators(db):
    # using Field(validation=) is the same as 'validators'
    assert db.b.a.requires[0].minsize == 5
    assert db.b.b.requires[0].minsize == 5


def test_is(db):
    assert isinstance(Consist.email.requires[0], isEmail)
    assert isinstance(Consist.url.requires[0], isUrl)
    assert isinstance(Consist.ip.requires[0], isIP)
    assert isinstance(Consist.image.requires[0], isImage)
    assert isinstance(Consist.emails.requires[0], isList)
    assert isinstance(Consist.emails.requires[0].children[0], isEmail)


def test_len(db):
    assert Len.a.requires[0].minsize == 5
    assert Len.a.requires[0].maxsize == 6
    assert Len.a.requires[0].inc[0] is True
    assert Len.a.requires[0].inc[1] is False
    assert Len.b.requires[0].minsize == 4
    assert Len.b.requires[0].maxsize == 13
    assert Len.b.requires[0].inc[0] is False
    assert Len.b.requires[0].inc[1] is False
    assert Len.c.requires[0].minsize == 5
    assert Len.c.requires[0].maxsize == 12
    assert Len.c.requires[0].inc[0] is True
    assert Len.c.requires[0].inc[1] is True
    assert Len.d.requires[0].minsize == 5
    assert Len.d.requires[0].maxsize == 13
    assert Len.d.requires[0].inc[0] is True
    assert Len.d.requires[0].inc[1] is False


def test_in(db):
    assert isinstance(Inside.a.requires[1], inSet)
    assert isinstance(Inside.b.requires[1], inRange)


def test_numerical(db):
    assert Num.a.requires[1].minimum == 0
    assert Num.b.requires[1].maximum == 5
    assert Num.a.requires[1].inc[0] is False
    assert Num.c.requires[1].minimum == 0
    assert Num.c.requires[1].maximum == 4
    assert Num.c.requires[1].inc[1] is True


def test_eq(db):
    assert isinstance(Eq.a.requires[1], Equals)
    assert isinstance(Eq.b.requires[1], Equals)
    assert isinstance(Eq.c.requires[1], Not)
    assert isinstance(Eq.c.requires[1].children[0], Equals)


def test_match(db):
    assert isinstance(Match.a.requires[1], Matches)
    assert isinstance(Match.b.requires[1], Matches)


def test_any(db):
    assert isinstance(Anyone.a.requires[1], Any)


def test_processors(db):
    assert isinstance(Proc.a.requires[1], Lower)
    assert isinstance(Proc.b.requires[1], Upper)
    assert isinstance(Proc.c.requires[1], Cleanup)
    assert isinstance(Proc.d.requires[1], Urlify)
    assert isinstance(Proc.e.requires[1], Crypt)
    assert isinstance(Proc.f.requires[1], Crypt)


def test_presence(db):
    assert len(Person.name.requires) == 2
    assert isinstance(Person.name.requires[0], isntEmpty)
    assert isinstance(Person.name.requires[1], hasLength)
    assert len(Person.surname.requires) == 2
    assert isinstance(Person.surname.requires[0], isntEmpty)
    assert isinstance(Person.surname.requires[1], hasLength)
    assert len(Thing.name.requires) == 2
    assert isinstance(Thing.name.requires[0], isntEmpty)
    assert isinstance(Thing.name.requires[1], hasLength)
    assert len(Thing.color.requires) == 2
    assert isinstance(Thing.color.requires[0], hasLength)
    assert isinstance(Thing.color.requires[1], inSet)
    assert isinstance(Thing.person.requires[0], inDB)


def test_unique(db):
    assert isinstance(Thing.uid.requires[2], notInDB)


def test_allow(db):
    assert isinstance(Allowed.a.requires[0], Allow)
    assert isinstance(Allowed.b.requires[0], isEmptyOr)
    assert isinstance(Allowed.c.requires[0], isEmptyOr)
    assert isinstance(Allowed.a.requires[0].children[1], inSet)
    assert isinstance(Allowed.b.requires[0].children[1], inSet)
    assert isinstance(Allowed.c.requires[0].children[1], inSet)


def test_validation(db):
    #: 'is'
    is_data = {
        'name': 'foo',
        'val': 1,
        'fval': 1.5,
        'text': 'Lorem ipsum',
        'password': 'notverysecret',
        'd': '{:%Y-%m-%d}'.format(datetime.utcnow()),
        't': '15:23',
        'dt': '2015-12-23T15:23:00',
        'json': '{}'
    }
    errors = A.validate(is_data)
    assert not errors
    d = dict(is_data)
    d['val'] = 'foo'
    errors = A.validate(d)
    assert 'val' in errors and len(errors) == 1
    d = dict(is_data)
    d['fval'] = 'bar'
    errors = A.validate(d)
    assert 'fval' in errors and len(errors) == 1
    d = dict(is_data)
    d['d'] = 'foo'
    errors = A.validate(d)
    assert 'd' in errors and len(errors) == 1
    d = dict(is_data)
    d['t'] = 'bar'
    errors = A.validate(d)
    assert 't' in errors and len(errors) == 1
    d = dict(is_data)
    d['dt'] = 'foo'
    errors = A.validate(d)
    assert 'dt' in errors and len(errors) == 1
    d = dict(is_data)
    d['json'] = 'bar'
    errors = A.validate(d)
    assert 'json' in errors and len(errors) == 1
    errors = Consist.validate({'email': 'foo'})
    assert 'email' in errors
    errors = Consist.validate({'url': 'notanurl'})
    assert 'url' in errors
    errors = Consist.validate({'url': 'http://domain.com/'})
    assert 'url' not in errors
    errors = Consist.validate({'url': 'http://domain.com'})
    assert 'url' not in errors
    errors = Consist.validate({'url': 'domain.com'})
    assert 'url' not in errors
    errors = Consist.validate({'ip': 'foo'})
    assert 'ip' in errors
    errors = Consist.validate({'emails': 'foo'})
    assert 'emails' in errors
    errors = Consist.validate({'emailsplit': 'foo'})
    assert 'emailsplit' in errors
    errors = Consist.validate({'emailsplit': '1@asd.com, 2@asd.com'})
    assert 'emailsplit' not in errors
    errors = Consist.validate({'emails': ['1@asd.com', '2@asd.com']})
    assert 'emails' not in errors
    #: 'len'
    len_data = {'a': '12345', 'b': '12345', 'c': '12345', 'd': '12345'}
    errors = Len.validate(len_data)
    assert not errors
    d = dict(len_data)
    d['a'] = 'ciao'
    errors = Len.validate(d)
    assert 'a' in errors and len(errors) == 1
    d = dict(len_data)
    d['b'] = 'ciao'
    errors = Len.validate(d)
    assert 'b' in errors and len(errors) == 1
    d = dict(len_data)
    d['b'] = '1234567890123'
    errors = Len.validate(d)
    assert 'b' in errors and len(errors) == 1
    d = dict(len_data)
    d['c'] = 'ciao'
    errors = Len.validate(d)
    assert 'c' in errors and len(errors) == 1
    d = dict(len_data)
    d['c'] = '1234567890123'
    errors = Len.validate(d)
    assert 'c' in errors and len(errors) == 1
    d = dict(len_data)
    d['d'] = 'ciao'
    errors = Len.validate(d)
    assert 'd' in errors and len(errors) == 1
    d = dict(len_data)
    d['d'] = '1234567890123'
    errors = Len.validate(d)
    assert 'd' in errors and len(errors) == 1
    #: 'in'
    in_data = {'a': 'a', 'b': 2}
    errors = Inside.validate(in_data)
    assert not errors
    d = dict(in_data)
    d['a'] = 'c'
    errors = Inside.validate(d)
    assert 'a' in errors and len(errors) == 1
    d = dict(in_data)
    d['b'] = 0
    errors = Inside.validate(d)
    assert 'b' in errors and len(errors) == 1
    d = dict(in_data)
    d['b'] = 7
    errors = Inside.validate(d)
    assert 'b' in errors and len(errors) == 1
    #: 'gt', 'lt', 'gte', 'lte'
    num_data = {'a': 1, 'b': 4, 'c': 2}
    errors = Num.validate(num_data)
    assert not errors
    d = dict(num_data)
    d['a'] = 0
    errors = Num.validate(d)
    assert 'a' in errors and len(errors) == 1
    d = dict(num_data)
    d['b'] = 5
    errors = Num.validate(d)
    assert 'b' in errors and len(errors) == 1
    d = dict(num_data)
    d['c'] = 0
    errors = Num.validate(d)
    assert 'c' in errors and len(errors) == 1
    d = dict(num_data)
    d['c'] = 5
    errors = Num.validate(d)
    assert 'c' in errors and len(errors) == 1
    #: 'equals'
    eq_data = {'a': 'asd', 'b': 2, 'c': 2.3}
    errors = Eq.validate(eq_data)
    assert not errors
    d = dict(eq_data)
    d['a'] = 'lol'
    errors = Eq.validate(d)
    assert 'a' in errors and len(errors) == 1
    d = dict(eq_data)
    d['b'] = 3
    errors = Eq.validate(d)
    assert 'b' in errors and len(errors) == 1
    #: 'not'
    d = dict(eq_data)
    d['c'] = 2.4
    errors = Eq.validate(d)
    assert 'c' in errors and len(errors) == 1
    #: 'match'
    match_data = {'a': 'abc', 'b': 'ab'}
    errors = Match.validate(match_data)
    assert not errors
    d = dict(match_data)
    d['a'] = 'lol'
    errors = Match.validate(d)
    assert 'a' in errors and len(errors) == 1
    d = dict(match_data)
    d['b'] = 'abc'
    errors = Match.validate(d)
    assert 'b' in errors and len(errors) == 1
    d['b'] = 'lol'
    errors = Match.validate(d)
    assert 'b' in errors and len(errors) == 1
    #: 'any'
    errors = Anyone.validate({'a': 'foo'})
    assert not errors
    errors = Anyone.validate({'a': 'walter@massivedynamics.com'})
    assert not errors
    errors = Anyone.validate({'a': 'lol'})
    assert 'a' in errors
    #: 'allow'
    allow_data = {'a': 'a', 'b': 'a', 'c': 'a'}
    errors = Allowed.validate(allow_data)
    assert not errors
    d = dict(allow_data)
    d['a'] = None
    errors = Allowed.validate(d)
    assert not errors
    d['a'] = 'foo'
    errors = Allowed.validate(d)
    assert 'a' in errors and len(errors) == 1
    d = dict(allow_data)
    d['b'] = ''
    errors = Allowed.validate(d)
    assert not errors
    d['b'] = None
    errors = Allowed.validate(d)
    assert not errors
    d['b'] = 'foo'
    errors = Allowed.validate(d)
    assert 'b' in errors and len(errors) == 1
    d = dict(allow_data)
    d['c'] = ''
    errors = Allowed.validate(d)
    assert not errors
    d['c'] = None
    errors = Allowed.validate(d)
    assert not errors
    d['c'] = 'foo'
    errors = Allowed.validate(d)
    assert 'c' in errors and len(errors) == 1
    #: processing validators
    assert Proc.a.validate('Asd')[0] == 'asd'
    assert Proc.b.validate('Asd')[0] == 'ASD'
    assert Proc.d.validate('Two Words')[0] == 'two-words'
    psw = str(Proc.e.validate('somepassword')[0])
    assert psw[:23] == 'pbkdf2(1000,20,sha512)$'
    assert psw[23:] != 'somepassword'
    psw = str(Proc.f.validate('somepassword')[0])
    assert psw[:4] == 'md5$'
    assert psw[4:] != 'somepassword'
    #: 'presence'
    mario = {'name': 'mario'}
    errors = Person.validate(mario)
    assert 'surname' in errors
    assert len(errors) == 1
    #: 'presence' with reference, 'unique'
    thing = {'name': 'a', 'person': 5, 'color': 'blue', 'uid': 'lol'}
    errors = Thing.validate(thing)
    assert 'person' in errors
    assert len(errors) == 1
    mario = {'name': 'mario', 'surname': 'draghi'}
    mario = Person.create(mario)
    assert len(mario.errors.keys()) == 0
    assert mario.id == 1
    thing = {'name': 'euro', 'person': mario.id, 'color': 'red', 'uid': 'lol'}
    thing = Thing.create(thing)
    assert len(thing.errors.keys()) == 0
    thing = {'name': 'euro2', 'person': mario.id, 'color': 'red', 'uid': 'lol'}
    errors = Thing.validate(thing)
    assert len(errors) == 1
    assert 'uid' in errors


def test_multi(db):
    p = db.Person(name="mario")
    base_data = {
        'date': '{0:%d/%m/%Y}'.format(datetime.utcnow()+timedelta(days=1)),
        'type': 'a',
        'inside': 'asd',
        'number': 1,
        'yep': 'asd',
        'psw': 'password',
        'person': p.id
    }
    #: everything ok
    res = Mixed.create(base_data)
    assert res.id == 1
    assert len(res.errors.keys()) == 0
    #: invalid belongs
    vals = dict(base_data)
    del vals['person']
    res = Mixed.create(vals)
    assert res.id is None
    assert len(res.errors.keys()) == 1
    assert 'person' in res.errors
    #: invalid date range
    vals = dict(base_data)
    vals['date'] = '{0:%d/%m/%Y}'.format(datetime.utcnow()-timedelta(days=2))
    res = Mixed.create(vals)
    assert res.id is None
    assert len(res.errors.keys()) == 1
    assert 'date' in res.errors
    #: invalid date format
    vals['date'] = '76-12-1249'
    res = Mixed.create(vals)
    assert res.id is None
    assert len(res.errors.keys()) == 1
    assert 'date' in res.errors
    #: invalid in
    vals = dict(base_data)
    vals['type'] = ' '
    res = Mixed.create(vals)
    assert res.id is None
    assert len(res.errors.keys()) == 1
    assert 'type' in res.errors
    #: empty number
    vals = dict(base_data)
    vals['number'] = None
    res = Mixed.create(vals)
    assert res.id == 2
    assert len(res.errors.keys()) == 0
    #: invalid number
    vals = dict(base_data)
    vals['number'] = 'asd'
    res = Mixed.create(vals)
    assert res.id is None
    assert len(res.errors.keys()) == 1
    assert 'number' in res.errors
    #: invalid empty
    vals = dict(base_data)
    vals['dont'] = '2'
    res = Mixed.create(vals)
    assert res.id is None
    assert len(res.errors.keys()) == 1
    assert 'dont' in res.errors
    #: invalid presence
    vals = dict(base_data)
    vals['yep'] = ''
    res = Mixed.create(vals)
    assert res.id is None
    assert len(res.errors.keys()) == 1
    assert 'yep' in res.errors
    #: invalid password
    vals = dict(base_data)
    vals['psw'] = ''
    res = Mixed.create(vals)
    assert res.id is None
    assert len(res.errors.keys()) == 1
    assert 'psw' in res.errors
    vals['psw'] = 'aksjfalsdkjflkasjdflkajsldkjfalslkdfjaslkdjf'
    res = Mixed.create(vals)
    assert res.id is None
    assert len(res.errors.keys()) == 1
    assert 'psw' in res.errors



# -*- coding: utf-8 -*-
"""
    emmett._internal
    ----------------

    Provides internally used helpers and objects.

    :copyright: 2014 Giovanni Barillari

    Several parts of this code comes from Flask and Werkzeug.
    :copyright: (c) 2014 by Armin Ronacher.

    :license: BSD-3-Clause
"""

from __future__ import annotations

import asyncio
import datetime
import os
import pkgutil
import sys
import traceback
import warnings

from functools import partial
from shutil import copyfileobj
from types import ModuleType
from typing import Any, Generic, Optional

import pendulum

from .typing import T


class ProxyMixin:
    def _get_robj(self):
        raise NotImplementedError

    def __getitem__(self, key):
        return self._get_robj()[key]

    def __setitem__(self, key, value):
        self._get_robj()[key] = value

    def __delitem__(self, key):
        del self._get_robj()[key]

    def __getattr__(self, name: str) -> Any:
        return getattr(self._get_robj(), name)

    def __setattr__(self, name, value):
        setattr(self._get_robj(), name, value)

    def __delattr__(self, name):
        delattr(self._get_robj(), name)

    def __bool__(self):
        try:
            return bool(self._get_robj())
        except RuntimeError:
            return False

    def __eq__(self, obj) -> bool:
        return self._get_robj() == obj

    def __ne__(self, obj) -> bool:
        return self._get_robj() != obj

    def __call__(self, *args, **kwargs):
        return self._get_robj()(*args, **kwargs)

    def __iter__(self):
        return iter(self._get_robj())

    def __contains__(self, element):
        return element in self._get_robj()

    def __dir__(self):
        try:
            return dir(self._get_robj())
        except RuntimeError:
            return []

    @property
    def __dict__(self):
        try:
            return self._get_robj().__dict__
        except RuntimeError:
            raise AttributeError('__dict__')

    def __str__(self):
        return str(self._get_robj())

    def __repr__(self):
        try:
            obj = self._get_robj()
        except RuntimeError:
            return '<%s unbound>' % self.__class__.__name__
        return repr(obj)


class ObjectProxy(ProxyMixin, Generic[T]):
    __slots__ = ('__obj', '__name__')

    def __init__(self, obj: Any, name: str):
        object.__setattr__(self, '_ObjectProxy__obj', obj)
        object.__setattr__(self, '__name__', name)

    def _get_robj(self) -> T:
        return getattr(self.__obj, self.__name__)


class ContextVarProxy(ProxyMixin, Generic[T]):
    __slots__ = ('__obj', '__name__')

    def __init__(self, obj: Any, name: str):
        object.__setattr__(self, '_ContextVarProxy__obj', obj)
        object.__setattr__(self, '__name__', name)

    def _get_robj(self) -> T:
        return getattr(self.__obj.get(), self.__name__)


class ImmutableListMixin:
    _hash_cache = None

    def __hash__(self) -> Optional[int]:  # type: ignore
        if self._hash_cache is not None:
            return self._hash_cache
        rv = self._hash_cache = hash(tuple(self))  # type: ignore
        return rv

    def __reduce_ex__(self, protocol):
        return type(self), (list(self),)

    def __delitem__(self, key):
        _is_immutable(self)

    def __iadd__(self, other):
        _is_immutable(self)

    def __imul__(self, other):
        _is_immutable(self)

    def __setitem__(self, key, value):
        _is_immutable(self)

    def append(self, item):
        _is_immutable(self)

    def remove(self, itme):
        _is_immutable(self)

    def extend(self, iterable):
        _is_immutable(self)

    def insert(self, pos, value):
        _is_immutable(self)

    def pop(self, index=-1):
        _is_immutable(self)

    def reverse(self):
        _is_immutable(self)

    def sort(self, cmp=None, key=None, reverse=None):
        _is_immutable(self)


class ImmutableList(ImmutableListMixin, list):  # type: ignore
    def __repr__(self):
        return '%s(%s)' % (
            self.__class__.__name__, list.__repr__(self)
        )


class LoopFileCtxWrapper:
    __slots__ = ('_coro', '_obj')

    def __init__(self, coro):
        self._coro = coro
        self._obj = None

    def __await__(self):
        return self._coro.__await__()

    async def __aenter__(self):
        self._obj = await self._coro
        return self._obj

    async def __aexit__(self, exc_type, exc, tb):
        await self._obj.close()
        self._obj = None


class LoopFileWrapper:
    __slots__ = ('_file', '_loop')

    def __init__(self, f, loop=None):
        self._file = f
        self._loop = loop or asyncio.get_running_loop()

    async def read(self, *args, **kwargs):
        return await self._loop.run_in_executor(
            None, partial(self._file.read, *args, **kwargs))

    async def write(self, *args, **kwargs):
        return await self._loop.run_in_executor(
            None, partial(self._file.write, *args, **kwargs))

    async def close(self, *args, **kwargs):
        return await self._loop.run_in_executor(
            None, partial(self._file.close, *args, **kwargs))

    def __getattr__(self, name):
        return getattr(self._file, name)


def _is_immutable(self):
    raise TypeError('%r objects are immutable' % self.__class__.__name__)


async def _loop_open_file(loop, *args, **kwargs):
    f = await loop.run_in_executor(None, partial(open, *args, **kwargs))
    return LoopFileWrapper(f, loop)


def loop_open_file(*args, **kwargs):
    return LoopFileCtxWrapper(
        _loop_open_file(asyncio.get_running_loop(), *args, **kwargs)
    )


async def loop_copyfileobj(fsrc, fdst, length=None):
    return await asyncio.get_running_loop().run_in_executor(
        None, partial(copyfileobj, fsrc, fdst, length)
    )


#: application loaders
def get_app_module(
    module_name: str,
    raise_on_failure: bool = True
) -> Optional[ModuleType]:
    try:
        __import__(module_name)
    except ImportError:
        if sys.exc_info()[-1].tb_next:
            raise RuntimeError(
                f"While importing '{module_name}', an ImportError was raised:"
                f"\n\n{traceback.format_exc()}"
            )
        elif raise_on_failure:
            raise RuntimeError(f"Could not import '{module_name}'.")
        else:
            return
    return sys.modules[module_name]


def find_best_app(module: ModuleType) -> Any:
    #: Given a module instance this tries to find the best possible
    #  application in the module.
    from .app import App # noqa

    # Search for the most common names first.
    for attr_name in ('app', 'application'):
        app = getattr(module, attr_name, None)
        if isinstance(app, App):
            return app

    # Otherwise find the only object that is an App instance.
    matches = [v for k, v in module.__dict__.items() if isinstance(v, App)]

    if len(matches) == 1:
        return matches[0]
    raise RuntimeError(
        f"Failed to find Emmett application in module '{module.__name__}'."
    )


def locate_app(module_name: str, app_name: str, raise_on_failure: bool = True) -> Any:
    module = get_app_module(module_name, raise_on_failure=raise_on_failure)
    if app_name:
        return getattr(module, app_name, None)
    return find_best_app(module)


#: deprecation helpers
class RemovedInNextVersionWarning(DeprecationWarning):
    pass


class deprecated(object):
    def __init__(self, old_method_name, new_method_name, class_name=None, s=0):
        self.class_name = class_name
        self.old_method_name = old_method_name
        self.new_method_name = new_method_name
        self.additional_stack = s

    def __call__(self, f):
        def wrapped(*args, **kwargs):
            warn_of_deprecation(
                self.old_method_name, self.new_method_name, self.class_name,
                3 + self.additional_stack)
            return f(*args, **kwargs)
        return wrapped


warnings.simplefilter('always', RemovedInNextVersionWarning)


def warn_of_deprecation(old_name, new_name, prefix=None, stack=2):
    msg = "%(old)s is deprecated, use %(new)s instead."
    if prefix:
        msg = "%(prefix)s." + msg
    warnings.warn(
        msg % {'old': old_name, 'new': new_name, 'prefix': prefix},
        RemovedInNextVersionWarning, stack)


#: app init helpers
def get_root_path(import_name):
    """Returns the path of the package or cwd if that cannot be found."""
    # Module already imported and has a file attribute.  Use that first.
    mod = sys.modules.get(import_name)
    if mod is not None and hasattr(mod, '__file__'):
        return os.path.dirname(os.path.abspath(mod.__file__))

    # Next attempt: check the loader.
    loader = pkgutil.get_loader(import_name)

    # Loader does not exist or we're referring to an unloaded main module
    # or a main module without path (interactive sessions), go with the
    # current working directory.
    if loader is None or import_name == '__main__':
        return os.getcwd()

    # For .egg, zipimporter does not have get_filename until Python 2.7.
    # Some other loaders might exhibit the same behavior.
    if hasattr(loader, 'get_filename'):
        filepath = loader.get_filename(import_name)
    else:
        # Fall back to imports.
        __import__(import_name)
        mod = sys.modules[import_name]
        filepath = getattr(mod, '__file__', None)

        # If we don't have a filepath it might be because we are a
        # namespace package.  In this case we pick the root path from the
        # first module that is contained in our package.
        if filepath is None:
            raise RuntimeError('No root path can be found for the provided '
                               'module "%s".  This can happen because the '
                               'module came from an import hook that does '
                               'not provide file name information or because '
                               'it\'s a namespace package.  In this case '
                               'the root path needs to be explicitly '
                               'provided.' % import_name)

    # filepath is import_name.py for a module, or __init__.py for a package.
    return os.path.dirname(os.path.abspath(filepath))


def create_missing_app_folders(app):
    try:
        for subfolder in ['languages', 'logs', 'static']:
            path = os.path.join(app.root_path, subfolder)
            if not os.path.exists(path):
                os.mkdir(path)
    except Exception:
        pass


#: monkey patches
def _pendulum_to_datetime(obj):
    return datetime.datetime(
        obj.year, obj.month, obj.day,
        obj.hour, obj.minute, obj.second, obj.microsecond,
        tzinfo=obj.tzinfo
    )


def _pendulum_to_naive_datetime(obj):
    obj = obj.in_timezone('UTC')
    return datetime.datetime(
        obj.year, obj.month, obj.day,
        obj.hour, obj.minute, obj.second, obj.microsecond
    )


def _pendulum_json(obj):
    return obj.for_json()


pendulum.DateTime.as_datetime = _pendulum_to_datetime  # type: ignore
pendulum.DateTime.as_naive_datetime = _pendulum_to_naive_datetime  # type: ignore
pendulum.DateTime.__json__ = _pendulum_json  # type: ignore



# -*- coding: utf-8 -*-
"""
    emmett.orm.migrations.scripts
    -----------------------------

    Provides scripts interface for migrations.

    :copyright: 2014 Giovanni Barillari

    Based on the code of Alembic (https://bitbucket.org/zzzeek/alembic)
    :copyright: (c) 2009-2015 by Michael Bayer

    :license: BSD-3-Clause
"""

import os
import re
import sys

from contextlib import contextmanager
from datetime import datetime
from renoir import Renoir

from ...html import asis
from .base import Migration
from .exceptions import (
    RangeNotAncestorError, MultipleHeads, ResolutionError, RevisionError
)
from .helpers import tuple_rev_as_scalar, format_with_comma
from .revisions import Revision, RevisionsMap


class ScriptDir(object):
    _slug_re = re.compile(r'\w+')
    _default_file_template = "%(rev)s_%(slug)s"

    def __init__(self, app, migrations_folder=None):
        self.app = app
        self.path = os.path.join(
            app.root_path, migrations_folder or 'migrations')
        if not os.path.exists(self.path):
            os.mkdir(self.path)
        self.cwd = os.path.dirname(__file__)
        self.file_template = self.app.config.migrations.file_template or \
            self._default_file_template
        self.truncate_slug_length = \
            self.app.config.migrations.filename_len or 40
        self.revision_map = RevisionsMap(self.app, self._load_revisions)
        self.templater = Renoir(path=self.cwd, mode='plain')

    def _load_revisions(self):
        sys.path.insert(0, self.path)
        for rev_file in os.listdir(os.path.abspath(self.path)):
            script = Script._from_filename(self, rev_file)
            if script is None:
                continue
            yield script

    @contextmanager
    def _catch_revision_errors(
            self,
            ancestor=None, multiple_heads=None, start=None, end=None,
            resolution=None):
        try:
            yield
        except RangeNotAncestorError as rna:
            if start is None:
                start = rna.lower
            if end is None:
                end = rna.upper
            if not ancestor:
                ancestor = (
                    "Requested range %(start)s:%(end)s does not refer to "
                    "ancestor/descendant revisions along the same branch"
                )
            ancestor = ancestor % {"start": start, "end": end}
            raise Exception(ancestor)
        except MultipleHeads as mh:
            if not multiple_heads:
                multiple_heads = (
                    "Multiple head revisions are present for given "
                    "argument '%(head_arg)s'; please "
                    "specify a specific target revision, "
                    "'<branchname>@%(head_arg)s' to "
                    "narrow to a specific head, or 'heads' for all heads")
            multiple_heads = multiple_heads % {
                "head_arg": end or mh.argument,
                "heads": str(mh.heads)
            }
            raise Exception(multiple_heads)
        except ResolutionError as re:
            if resolution is None:
                resolution = "Can't locate revision identified by '%s'" % (
                    re.argument
                )
            raise Exception(resolution)
        except RevisionError as err:
            raise Exception(err.args[0])

    def walk_revisions(self, base="base", head="heads"):
        with self._catch_revision_errors(start=base, end=head):
            for rev in self.revision_map.iterate_revisions(
                    head, base, inclusive=True):
                yield rev

    def get_revision(self, revid):
        with self._catch_revision_errors():
            return self.revision_map.get_revision(revid)

    def get_revisions(self, revid):
        with self._catch_revision_errors():
            return self.revision_map.get_revisions(revid)

    def get_upgrade_revs(self, destination, current_rev):
        with self._catch_revision_errors(
                ancestor="Destination %(end)s is not a valid upgrade "
                "target from current head(s)", end=destination):
            revs = self.revision_map.iterate_revisions(
                destination, current_rev, implicit_base=True)
            return reversed(list(revs))

    def get_downgrade_revs(self, destination, current_rev):
        with self._catch_revision_errors(
                ancestor="Destination %(end)s is not a valid downgrade "
                "target from current head(s)", end=destination):
            revs = self.revision_map.iterate_revisions(
                current_rev, destination)
            return list(revs)

    def _rev_filename(self, revid, message, creation_date):
        slug = "_".join(self._slug_re.findall(message or "")).lower()
        if len(slug) > self.truncate_slug_length:
            slug = slug[:self.truncate_slug_length].rsplit('_', 1)[0] + '_'
        filename = "%s.py" % (
            self.file_template % {
                'rev': revid,
                'slug': slug,
                'year': creation_date.year,
                'month': creation_date.month,
                'day': creation_date.day,
                'hour': creation_date.hour,
                'minute': creation_date.minute,
                'second': creation_date.second
            }
        )
        return filename

    def _generate_template(self, filename, ctx):
        rendered = self.templater.render('migration.tmpl', ctx)
        with open(os.path.join(self.path, filename), 'w') as f:
            f.write(rendered)

    def generate_revision(self, revid, message, head=None, splice=False, **kw):
        """Generate a new revision file.

        This runs the templater, and creates a new file.

        :param revid: String revision id.
        :param message: the revision message.
        :param head: the head revision to generate against. Defaults
         to the current "head" if no branches are present, else raises
         an exception.
        :param splice: if True, allow the "head" version to not be an
         actual head; otherwise, the selected head must be a head
         (e.g. endpoint) revision.
        """
        if head is None:
            head = "head"

        with self._catch_revision_errors(multiple_heads=(
            "Multiple heads are present; please specify the head "
            "revision on which the new revision should be based, "
            "or perform a merge."
        )):
            heads = self.revision_map.get_revisions(head)

        if len(set(heads)) != len(heads):
            raise Exception("Duplicate head revisions specified")

        creation_date = datetime.now()

        rev_filename = self._rev_filename(revid, message, creation_date)

        if not splice:
            for head in heads:
                if head is not None and not head.is_head:
                    raise Exception(
                        "Revision %s is not a head revision; please specify "
                        "--splice to create a new branch from this revision"
                        % head.revision)

        down_migration = tuple(
            h.revision if h is not None else None for h in heads)

        down_migration_var = tuple_rev_as_scalar(down_migration)
        if isinstance(down_migration_var, str):
            down_migration_var = "%r" % down_migration_var
        else:
            down_migration_var = str(down_migration_var)

        template_ctx = dict(
            asis=asis,
            up_migration=revid,
            down_migration=down_migration_var,
            creation_date=creation_date,
            down_migration_str=", ".join(r for r in down_migration),
            message=message if message is not None else ("empty message"),
            upgrades=kw.get('upgrades', ['pass']),
            downgrades=kw.get('downgrades', ['pass'])
        )
        self._generate_template(rev_filename, template_ctx)

        script = Script._from_filename(self, rev_filename)
        self.revision_map.add_revision(script)
        return script


class Script(Revision):
    _only_source_rev_file = re.compile(r'(?!__init__)(.*\.py)$')
    migration_class = None
    path = None

    def __init__(self, module, migration_class, path):
        self.module = module
        self.migration_class = migration_class
        self.path = path
        super(Script, self).__init__(
            self.migration_class.revision,
            self.migration_class.revises
        )

    @property
    def doc(self):
        return re.split("\n\n", self.longdoc)[0]

    @property
    def longdoc(self):
        doc = self.module.__doc__
        return doc.strip() if doc else ""

    @property
    def log_entry(self):
        entry = "Rev: %s%s%s%s\n" % (
            self.revision,
            " (head)" if self.is_head else "",
            " (branchpoint)" if self.is_branch_point else "",
            " (mergepoint)" if self.is_merge_point else "",
        )
        if self.is_merge_point:
            entry += "Merges: %s\n" % (self._format_down_revision(), )
        else:
            entry += "Parent: %s\n" % (self._format_down_revision(), )

        if self.is_branch_point:
            entry += "Branches into: %s\n" % (
                format_with_comma(self.nextrev))

        entry += "Path: %s\n" % (self.path,)

        entry += "\n%s\n" % (
            "\n".join(
                "    %s" % para
                for para in self.longdoc.splitlines()
            )
        )
        return entry

    def __str__(self):
        return "%s -> %s%s%s%s, %s" % (
            self._format_down_revision(),
            self.revision,
            " (head)" if self.is_head else "",
            " (branchpoint)" if self.is_branch_point else "",
            " (mergepoint)" if self.is_merge_point else "",
            self.doc)

    def _head_only(
            self, include_doc=False,
            include_parents=False, tree_indicators=True,
            head_indicators=True):
        text = self.revision
        if include_parents:
            text = "%s -> %s" % (
                self._format_down_revision(), text)
        if head_indicators or tree_indicators:
            text += "%s%s" % (
                " (head)" if self._is_real_head else "",
                " (effective head)" if self.is_head and
                    not self._is_real_head else ""
            )
        if tree_indicators:
            text += "%s%s" % (
                " (branchpoint)" if self.is_branch_point else "",
                " (mergepoint)" if self.is_merge_point else "",
            )
        if include_doc:
            text += ", %s" % self.doc
        return text

    def cmd_format(
        self,
            verbose,
            include_doc=False,
            include_parents=False, tree_indicators=True):
        if verbose:
            return self.log_entry
        else:
            return self._head_only(
                include_doc,
                include_parents, tree_indicators)

    def _format_down_revision(self):
        if not self.down_revision:
            return "<base>"
        else:
            return format_with_comma(self._versioned_down_revisions)

    @classmethod
    def _from_filename(cls, scriptdir, filename):
        py_match = cls._only_source_rev_file.match(filename)
        if not py_match:
            return None
        py_filename = py_match.group(1)
        py_module = py_filename.split('.py')[0]
        __import__(py_module)
        module = sys.modules[py_module]
        migration_class = getattr(module, 'Migration', None)
        if migration_class is None:
            for v in module.__dict__.values():
                if isinstance(v, Migration):
                    migration_class = v
                    break
        return Script(
            module, migration_class, os.path.join(scriptdir.path, filename))



# -*- coding: utf-8 -*-
"""
    emmett.asgi.protocols.ws.auto
    -----------------------------

    Provides websocket auto protocol loader

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from . import protocols


if "websockets" in protocols:
    protocols.register("auto")(protocols.get("websockets"))
else:
    protocols.register("auto")(protocols.get("wsproto"))



# -*- coding: utf-8 -*-
"""
    emmett.templating.lexers
    ------------------------

    Provides the Emmett lexers for Renoir engine.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from renoir import Lexer

from ..ctx import current
from ..routing.urls import url


class HelpersLexer(Lexer):
    helpers = [
        '<script type="text/javascript" ' +
        'src="{}/__emmett__/jquery.min.js"></script>',
        '<script type="text/javascript" ' +
        'src="{}/__emmett__/helpers.js"></script>'
    ]

    def process(self, ctx, value):
        for helper in self.helpers:
            ctx.html(helper.format(current.app._router_http._prefix_main))


class MetaLexer(Lexer):
    def process(self, ctx, value):
        ctx.python_node('for name, value in current.response._meta_tmpl():')
        ctx.variable(
            "'<meta name=\"%s\" content=\"%s\" />' % (name, value)",
            escape=False)
        ctx.python_node('pass')
        ctx.python_node(
            'for name, value in current.response._meta_tmpl_prop():')
        ctx.variable(
            "'<meta property=\"%s\" content=\"%s\" />' % (name, value)",
            escape=False)
        ctx.python_node('pass')


class StaticLexer(Lexer):
    evaluate = True

    def process(self, ctx, value):
        file_name = value.split("?")[0]
        surl = url('static', file_name)
        file_ext = file_name.rsplit(".", 1)[-1]
        if file_ext == 'js':
            s = u'<script type="text/javascript" src="%s"></script>' % surl
        elif file_ext == "css":
            s = u'<link rel="stylesheet" href="%s" type="text/css" />' % surl
        else:
            s = None
        if s:
            ctx.html(s)


lexers = {
    'include_helpers': HelpersLexer(),
    'include_meta': MetaLexer(),
    'include_static': StaticLexer()
}



# -*- coding: utf-8 -*-
"""
    emmett.http
    -----------

    Provides the HTTP interfaces.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from __future__ import annotations

import errno
import os
import stat

from email.utils import formatdate
from hashlib import md5
from typing import Any, BinaryIO, Dict, Generator, Tuple

from ._internal import loop_open_file
from .ctx import current
from .libs.contenttype import contenttype

status_codes = {
    100: '100 CONTINUE',
    101: '101 SWITCHING PROTOCOLS',
    200: '200 OK',
    201: '201 CREATED',
    202: '202 ACCEPTED',
    203: '203 NON-AUTHORITATIVE INFORMATION',
    204: '204 NO CONTENT',
    205: '205 RESET CONTENT',
    206: '206 PARTIAL CONTENT',
    207: '207 MULTI-STATUS',
    300: '300 MULTIPLE CHOICES',
    301: '301 MOVED PERMANENTLY',
    302: '302 FOUND',
    303: '303 SEE OTHER',
    304: '304 NOT MODIFIED',
    305: '305 USE PROXY',
    307: '307 TEMPORARY REDIRECT',
    400: '400 BAD REQUEST',
    401: '401 UNAUTHORIZED',
    403: '403 FORBIDDEN',
    404: '404 NOT FOUND',
    405: '405 METHOD NOT ALLOWED',
    406: '406 NOT ACCEPTABLE',
    407: '407 PROXY AUTHENTICATION REQUIRED',
    408: '408 REQUEST TIMEOUT',
    409: '409 CONFLICT',
    410: '410 GONE',
    411: '411 LENGTH REQUIRED',
    412: '412 PRECONDITION FAILED',
    413: '413 REQUEST ENTITY TOO LARGE',
    414: '414 REQUEST-URI TOO LONG',
    415: '415 UNSUPPORTED MEDIA TYPE',
    416: '416 REQUESTED RANGE NOT SATISFIABLE',
    417: '417 EXPECTATION FAILED',
    422: '422 UNPROCESSABLE ENTITY',
    500: '500 INTERNAL SERVER ERROR',
    501: '501 NOT IMPLEMENTED',
    502: '502 BAD GATEWAY',
    503: '503 SERVICE UNAVAILABLE',
    504: '504 GATEWAY TIMEOUT',
    505: '505 HTTP VERSION NOT SUPPORTED',
}


class HTTPResponse(Exception):
    def __init__(
        self,
        status_code: int,
        *,
        headers: Dict[str, str] = {'content-type': 'text/plain'},
        cookies: Dict[str, Any] = {}
    ):
        self.status_code: int = status_code
        self._headers: Dict[str, str] = headers
        self._cookies: Dict[str, Any] = cookies

    @property
    def headers(self) -> Generator[Tuple[bytes, bytes], None, None]:
        for key, val in self._headers.items():
            yield key.encode('latin-1'), val.encode('latin-1')
        for cookie in self._cookies.values():
            yield b'set-cookie', str(cookie)[12:].encode('latin-1')

    async def _send_headers(self, send):
        await send({
            'type': 'http.response.start',
            'status': self.status_code,
            'headers': self.headers
        })

    async def _send_body(self, send):
        await send({'type': 'http.response.body'})

    async def send(self, scope, send):
        await self._send_headers(send)
        await self._send_body(send)


class HTTPBytes(HTTPResponse):
    def __init__(
        self,
        status_code: int,
        body: bytes = b'',
        headers: Dict[str, str] = {'content-type': 'text/plain'},
        cookies: Dict[str, Any] = {}
    ):
        super().__init__(status_code, headers=headers, cookies=cookies)
        self.body = body

    async def _send_body(self, send):
        await send({
            'type': 'http.response.body',
            'body': self.body,
            'more_body': False
        })


class HTTP(HTTPResponse):
    def __init__(
        self,
        status_code: int,
        body: str = '',
        headers: Dict[str, str] = {'content-type': 'text/plain'},
        cookies: Dict[str, Any] = {}
    ):
        super().__init__(status_code, headers=headers, cookies=cookies)
        self.body = body

    @property
    def encoded_body(self):
        return self.body.encode('utf-8')

    async def _send_body(self, send):
        await send({
            'type': 'http.response.body',
            'body': self.encoded_body,
            'more_body': False
        })


class HTTPRedirect(HTTPResponse):
    def __init__(
        self,
        status_code: int,
        location: str,
        cookies: Dict[str, Any] = {}
    ):
        location = location.replace('\r', '%0D').replace('\n', '%0A')
        super().__init__(
            status_code,
            headers={'location': location},
            cookies=cookies
        )


class HTTPFile(HTTPResponse):
    def __init__(
        self,
        file_path: str,
        headers: Dict[str, str] = {},
        cookies: Dict[str, Any] = {},
        chunk_size: int = 4096
    ):
        super().__init__(200, headers=headers, cookies=cookies)
        self.file_path = file_path
        self.chunk_size = chunk_size

    def _get_stat_headers(self, stat_data):
        content_length = str(stat_data.st_size)
        last_modified = formatdate(stat_data.st_mtime, usegmt=True)
        etag_base = str(stat_data.st_mtime) + '_' + str(stat_data.st_size)
        etag = md5(etag_base.encode('utf-8')).hexdigest()
        return {
            'content-type': contenttype(self.file_path),
            'content-length': content_length,
            'last-modified': last_modified,
            'etag': etag
        }

    async def send(self, scope, send):
        try:
            stat_data = os.stat(self.file_path)
            if not stat.S_ISREG(stat_data.st_mode):
                await HTTP(403).send(scope, send)
                return
            self._headers.update(self._get_stat_headers(stat_data))
            await self._send_headers(send)
            await self._send_body(send)
        except IOError as e:
            if e.errno == errno.EACCES:
                await HTTP(403).send(scope, send)
            else:
                await HTTP(404).send(scope, send)

    async def _send_body(self, send):
        async with loop_open_file(self.file_path, mode='rb') as f:
            more_body = True
            while more_body:
                chunk = await f.read(self.chunk_size)
                more_body = len(chunk) == self.chunk_size
                await send({
                    'type': 'http.response.body',
                    'body': chunk,
                    'more_body': more_body,
                })


class HTTPIO(HTTPResponse):
    def __init__(
        self,
        io_stream: BinaryIO,
        headers: Dict[str, str] = {},
        cookies: Dict[str, Any] = {},
        chunk_size: int = 4096
    ):
        super().__init__(200, headers=headers, cookies=cookies)
        self.io_stream = io_stream
        self.chunk_size = chunk_size

    def _get_io_headers(self):
        content_length = str(self.io_stream.getbuffer().nbytes)
        return {
            'content-length': content_length
        }

    async def send(self, scope, send):
        self._headers.update(self._get_io_headers())
        await self._send_headers(send)
        await self._send_body(send)

    async def _send_body(self, send):
        more_body = True
        while more_body:
            chunk = await self.io_stream.read(self.chunk_size)
            more_body = len(chunk) == self.chunk_size
            await send({
                'type': 'http.response.body',
                'body': chunk,
                'more_body': more_body,
            })


def redirect(location: str, status_code: int = 303):
    response = current.response
    response.status = status_code
    raise HTTPRedirect(status_code, location, response.cookies)



# -*- coding: utf-8 -*-
"""
    emmett.testing.helpers
    ----------------------

    Provides helpers for testing suite.

    :copyright: 2014 Giovanni Barillari

    Several parts of this code comes from Werkzeug.
    :copyright: (c) 2015 by Armin Ronacher.

    :license: BSD-3-Clause
"""

import codecs
import mimetypes
import re
import sys

from http.cookiejar import CookieJar
from io import BytesIO
from urllib.request import Request as U2Request

# TODO: check conversions
from .._shortcuts import to_bytes, to_unicode
from ..datastructures import sdict
from .urls import get_host, uri_to_iri, url_quote

_quoted_string_re = r'"[^"\\]*(?:\\.[^"\\]*)*"'
_option_header_piece_re = re.compile(
    r';\s*(%s|[^\s;,=]+)\s*(?:=\s*(%s|[^;,]+)?)?\s*' %
    (_quoted_string_re, _quoted_string_re)
)
_option_header_start_mime_type = re.compile(r',\s*([^;,\s]+)([;,]\s*.+)?')


class _TestCookieHeaders(object):
    def __init__(self, headers):
        self.headers = headers

    def getheaders(self, name):
        headers = []
        name = name.lower()
        for k, v in self.headers:
            if k.lower() == name:
                headers.append(v)
        return headers

    def get_all(self, name, default=None):
        rv = []
        for k, v in self.headers:
            if k.lower() == name.lower():
                rv.append(v)
        return rv or default or []


class _TestCookieResponse(object):
    def __init__(self, headers):
        self.headers = _TestCookieHeaders(headers)

    def info(self):
        return self.headers


class TestCookieJar(CookieJar):
    def inject_asgi(self, scope):
        cvals = []
        for cookie in self:
            cvals.append('%s=%s' % (cookie.name, cookie.value))
        if cvals:
            scope['headers'].append(
                (b'cookie', '; '.join(cvals).encode('utf-8')))

    def extract_asgi(self, scope, headers):
        self.extract_cookies(
            _TestCookieResponse(headers),
            U2Request(get_current_url(scope)),
        )


class Headers(dict):
    def __init__(self, headers=[]):
        super(Headers, self).__init__()
        self._list = []
        for header in headers:
            key, value = header[0].lower(), header[1]
            self._list.append((key, value))
            self[key] = value

    def __getitem__(self, name):
        return super(Headers, self).__getitem__(name.lower())

    def __setitem__(self, name, value):
        return super(Headers, self).__setitem__(name.lower(), value)

    def __iter__(self):
        return iter(self._list)

    def get(self, name, d=None, type=None):
        rv = super(Headers, self).get(name.lower)
        if rv is None:
            return d
        if type is None:
            return rv
        try:
            rv = type(rv)
        except ValueError:
            pass
        return rv

    def to_asgi_list(self):
        return list(self)


class _FileHandler(object):
    def __init__(self, stream=None, filename=None, name=None,
                 content_type=None, content_length=None,
                 headers=None):
        self.name = name
        self.stream = stream or BytesIO()

        # if no filename is provided we can attempt to get the filename
        # from the stream object passed.  There we have to be careful to
        # skip things like <fdopen>, <stderr> etc.  Python marks these
        # special filenames with angular brackets.
        if filename is None:
            filename = getattr(stream, 'name', None)
            #s = make_literal_wrapper(filename)
            if filename and filename[0] == '<' and filename[-1] == '>':
                filename = None

            # On Python 3 we want to make sure the filename is always unicode.
            # This might not be if the name attribute is bytes due to the
            # file being opened from the bytes API.
            if isinstance(filename, bytes):
                filename = filename.decode(get_filesystem_encoding(),
                                           'replace')

        self.filename = filename
        if headers is None:
            headers = Headers()
        self.headers = headers
        if content_type is not None:
            headers['Content-Type'] = content_type
        if content_length is not None:
            headers['Content-Length'] = str(content_length)

    def _parse_content_type(self):
        if not hasattr(self, '_parsed_content_type'):
            self._parsed_content_type = \
                parse_options_header(self.content_type)

    @property
    def content_type(self):
        return self.headers.get('content-type')

    @property
    def content_length(self):
        return int(self.headers.get('content-length') or 0)

    @property
    def mimetype(self):
        self._parse_content_type()
        return self._parsed_content_type[0].lower()

    @property
    def mimetype_params(self):
        self._parse_content_type()
        return self._parsed_content_type[1]

    def save(self, dst, buffer_size=16384):
        """Save the file to a destination path or file object.  If the
        destination is a file object you have to close it yourself after the
        call.  The buffer size is the number of bytes held in memory during
        the copy process.  It defaults to 16KB.
        For secure file saving also have a look at :func:`secure_filename`.
        :param dst: a filename or open file object the uploaded file
                    is saved to.
        :param buffer_size: the size of the buffer.  This works the same as
                            the `length` parameter of
                            :func:`shutil.copyfileobj`.
        """
        from shutil import copyfileobj
        close_dst = False
        if isinstance(dst, str):
            dst = open(dst, 'wb')
            close_dst = True
        try:
            copyfileobj(self.stream, dst, buffer_size)
        finally:
            if close_dst:
                dst.close()

    def close(self):
        """Close the underlying file if possible."""
        try:
            self.stream.close()
        except Exception:
            pass

    def __nonzero__(self):
        return bool(self.filename)
    __bool__ = __nonzero__

    def __getattr__(self, name):
        return getattr(self.stream, name)

    def __iter__(self):
        return iter(self.readline, '')

    def __repr__(self):
        return '<%s: %r (%r)>' % (
            self.__class__.__name__,
            self.filename,
            self.content_type
        )


class filesdict(sdict):
    def add_file(self, name, file, filename=None, content_type=None):
        if isinstance(file, _FileHandler):
            value = file
        else:
            if isinstance(file, str):
                if filename is None:
                    filename = file
                file = open(file, 'rb')
            if filename and content_type is None:
                content_type = mimetypes.guess_type(filename)[0] or \
                    'application/octet-stream'
            value = _FileHandler(file, filename, name, content_type)
        self[name] = value


def get_current_url(
    scope, root_only=False, strip_querystring=False, host_only=False,
    trusted_hosts=None
):
    headers = Headers(scope['headers'])
    components = [
        scope['scheme'], '://', get_host(scope, headers, trusted_hosts)]
    if host_only:
        return uri_to_iri(''.join(components) + '/')
    components.extend([
        url_quote(to_bytes(scope['root_path'])).rstrip('/'), '/'])
    if not root_only:
        components.append(url_quote(to_bytes(scope['path']).lstrip(b'/')))
        if not strip_querystring:
            if scope['query_string']:
                components.extend(['?', to_unicode(scope['query_string'])])
    return uri_to_iri(''.join(components))


def _is_ascii_encoding(encoding):
    if encoding is None:
        return False
    try:
        return codecs.lookup(encoding).name == 'ascii'
    except LookupError:
        return False


def get_filesystem_encoding():
    rv = sys.getfilesystemencoding()
    if (sys.platform.startswith('linux') or 'bsd' in sys.platform) and not rv \
       or _is_ascii_encoding(rv):
        return 'utf-8'
    return rv


def unquote_header_value(value, is_filename=False):
    if value and value[0] == value[-1] == '"':
        # this is not the real unquoting, but fixing this so that the
        # RFC is met will result in bugs with internet explorer and
        # probably some other browsers as well.  IE for example is
        # uploading files with "C:\foo\bar.txt" as filename
        value = value[1:-1]

        # if this is a filename and the starting characters look like
        # a UNC path, then just return the value without quotes.  Using the
        # replace sequence below on a UNC path has the effect of turning
        # the leading double slash into a single slash and then
        # _fix_ie_filename() doesn't work correctly.  See #458.
        if not is_filename or value[:2] != '\\\\':
            return value.replace('\\\\', '\\').replace('\\"', '"')
    return value


def parse_options_header(value, multiple=False):
    if not value:
        return '', {}
    result = []
    value = "," + value.replace("\n", ",")
    while value:
        match = _option_header_start_mime_type.match(value)
        if not match:
            break
        result.append(match.group(1))  # mimetype
        options = {}
        # Parse options
        rest = match.group(2)
        while rest:
            optmatch = _option_header_piece_re.match(rest)
            if not optmatch:
                break
            option, option_value = optmatch.groups()
            option = unquote_header_value(option)
            if option_value is not None:
                option_value = unquote_header_value(
                    option_value,
                    option == 'filename')
            options[option] = option_value
            rest = rest[optmatch.end():]
        result.append(options)
        if multiple is False:
            return tuple(result)
        value = rest

    return tuple(result)


def stream_encode_multipart(values, threshold=1024 * 500, boundary=None,
                            charset='utf-8'):
    """Encode a dict of values (either strings or file descriptors or
    :class:`FileStorage` objects.) into a multipart encoded string stored
    in a file descriptor.
    """
    if boundary is None:
        from time import time
        from random import random
        boundary = '---------------EmmettFormPart_%s%s' % (time(), random())
    _closure = [BytesIO(), 0, False]

    write_binary = _closure[0].write

    def write(string):
        write_binary(string.encode(charset))

    for key, values in values.items():
        if not isinstance(values, list):
            values = [values]
        for value in values:
            write('--%s\r\nContent-Disposition: form-data; name="%s"' %
                  (boundary, key))
            reader = getattr(value, 'read', None)
            if reader is not None:
                filename = getattr(value, 'filename',
                                   getattr(value, 'name', None))
                content_type = getattr(value, 'content_type', None)
                if content_type is None:
                    content_type = filename and \
                        mimetypes.guess_type(filename)[0] or \
                        'application/octet-stream'
                if filename is not None:
                    write('; filename="%s"\r\n' % filename)
                else:
                    write('\r\n')
                write('Content-Type: %s\r\n\r\n' % content_type)
                while 1:
                    chunk = reader(16384)
                    if not chunk:
                        break
                    write_binary(chunk)
            else:
                if not isinstance(value, str):
                    value = str(value)
                else:
                    value = to_bytes(value, charset)
                write('\r\n\r\n')
                write_binary(value)
            write('\r\n')
    write('--%s--\r\n' % boundary)

    length = int(_closure[0].tell())
    _closure[0].seek(0)
    return _closure[0], length, boundary



# -*- coding: utf-8 -*-
"""
    emmett.testing.env
    ------------------

    Provides environment class for testing suite.

    :copyright: 2014 Giovanni Barillari

    Several parts of this code comes from Werkzeug.
    :copyright: (c) 2015 by Armin Ronacher.

    :license: BSD-3-Clause
"""

import cgi
import sys

from io import BytesIO

from ..datastructures import sdict
from .helpers import Headers, filesdict, stream_encode_multipart
from .urls import iri_to_uri, url_fix, url_parse, url_unparse, url_encode


class ScopeBuilder(object):
    #: This class creates an ASGI environment for testing purposes.

    #: the server protocol to use.  defaults to HTTP/1.1
    server_protocol = '1.1'

    def __init__(
        self, path='/', base_url=None, query_string=None, method='GET',
        input_stream=None, content_type=None, content_length=None,
        errors_stream=None, headers=None, data=None, environ_base=None,
        environ_overrides=None, charset='utf-8'
    ):
        if query_string is None and '?' in path:
            path, query_string = path.split('?', 1)
        self.charset = charset
        self.path = iri_to_uri(path)
        if base_url is not None:
            base_url = url_fix(iri_to_uri(base_url, charset), charset)
        self.base_url = base_url
        if isinstance(query_string, (bytes, str)):
            self.query_string = query_string
        else:
            if query_string is None:
                query_string = sdict()
            elif not isinstance(query_string, dict):
                query_string = self._parse_querystring(query_string)
            self.args = query_string
        self.method = method
        if headers is None:
            headers = Headers()
        elif not isinstance(headers, Headers):
            headers = Headers(headers)
        self.headers = headers
        if content_type is not None:
            self.content_type = content_type
        if errors_stream is None:
            errors_stream = sys.stderr
        self.errors_stream = errors_stream
        self.environ_base = environ_base
        self.environ_overrides = environ_overrides
        self.input_stream = input_stream
        self.content_length = content_length
        self.closed = False

        if data:
            if input_stream is not None:
                raise TypeError('can\'t provide input stream and data')
            if isinstance(data, str):
                data = data.encode(self.charset)
            if isinstance(data, bytes):
                self.input_stream = BytesIO(data)
                if self.content_length is None:
                    self.content_length = len(data)
            else:
                for key, values in data.items():
                    if not isinstance(values, list):
                        values = [values]
                    for v in values:
                        if isinstance(v, (tuple)) or hasattr(v, 'read'):
                            self._add_file_from_data(key, v)
                        else:
                            if self.form[key] is None:
                                self.form[key] = []
                            self.form[key].append(v)

    @staticmethod
    def _parse_querystring(query_string):
        dget = cgi.parse_qs(query_string, keep_blank_values=1)
        params = sdict(dget)
        for key, value in params.items():
            if isinstance(value, list) and len(value) == 1:
                params[key] = value[0]
        return params

    def _add_file_from_data(self, key, value):
        if isinstance(value, tuple):
            self.files.add_file(key, *value)
        else:
            self.files.add_file(key, value)

    def _get_base_url(self):
        return url_unparse(
            (self.url_scheme, self.host, self.script_root, '', '')
        ).rstrip('/') + '/'

    def _set_base_url(self, value):
        if value is None:
            scheme = 'http'
            netloc = 'localhost'
            script_root = ''
        else:
            scheme, netloc, script_root, qs, anchor = url_parse(value)
            if qs or anchor:
                raise ValueError(
                    'base url must not contain a query string or fragment')
        self.script_root = script_root.rstrip('/')
        self.host = netloc
        self.url_scheme = scheme

    base_url = property(_get_base_url, _set_base_url)
    del _get_base_url, _set_base_url

    def _get_content_type(self):
        ct = self.headers.get('Content-Type')
        if ct is None and not self._input_stream:
            if self._files:
                return 'multipart/form-data'
            elif self._form:
                return 'application/x-www-form-urlencoded'
            return None
        return ct

    def _set_content_type(self, value):
        if value is None:
            self.headers.pop('Content-Type', None)
        else:
            self.headers['Content-Type'] = value

    content_type = property(_get_content_type, _set_content_type)
    del _get_content_type, _set_content_type

    def _get_content_length(self):
        return self.headers.get('Content-Length', type=int)

    def _set_content_length(self, value):
        if value is None:
            self.headers.pop('Content-Length', None)
        else:
            self.headers['Content-Length'] = str(value)

    content_length = property(_get_content_length, _set_content_length)
    del _get_content_length, _set_content_length

    def form_property(name, storage):
        key = '_' + name

        def getter(self):
            if self._input_stream is not None:
                raise AttributeError('an input stream is defined')
            rv = getattr(self, key)
            if rv is None:
                rv = storage()
                setattr(self, key, rv)

            return rv

        def setter(self, value):
            self._input_stream = None
            setattr(self, key, value)
        return property(getter, setter)

    form = form_property('form', sdict)
    files = form_property('files', filesdict)
    del form_property

    def _get_input_stream(self):
        return self._input_stream

    def _set_input_stream(self, value):
        self._input_stream = value
        self._form = self._files = None

    input_stream = property(_get_input_stream, _set_input_stream, doc='''
        An optional input stream.  If you set this it will clear
        :attr:`form` and :attr:`files`.''')
    del _get_input_stream, _set_input_stream

    def _get_query_string(self):
        if self._query_string is None:
            if self._args is not None:
                return url_encode(self._args, charset=self.charset)
            return ''
        return self._query_string

    def _set_query_string(self, value):
        self._query_string = value
        self._args = None

    query_string = property(_get_query_string, _set_query_string)
    del _get_query_string, _set_query_string

    def _get_args(self):
        if self._query_string is not None:
            raise AttributeError('a query string is defined')
        if self._args is None:
            self._args = sdict()
        return self._args

    def _set_args(self, value):
        self._query_string = None
        self._args = value

    args = property(_get_args, _set_args)
    del _get_args, _set_args

    @property
    def server_name(self):
        """The server name (read-only, use :attr:`host` to set)"""
        return self.host.split(':', 1)[0]

    @property
    def server_port(self):
        """The server port as integer (read-only, use :attr:`host` to set)"""
        pieces = self.host.split(':', 1)
        if len(pieces) == 2 and pieces[1].isdigit():
            return int(pieces[1])
        elif self.url_scheme == 'https':
            return 443
        return 80

    def __del__(self):
        try:
            self.close()
        except Exception:
            pass

    def close(self):
        """Closes all files.  If you put real :class:`file` objects into the
        :attr:`files` dict you can call this method to automatically close
        them all in one go.
        """
        if self.closed:
            return
        try:
            files = self.files.values()
        except AttributeError:
            files = ()
        for f in files:
            try:
                f.close()
            except Exception:
                pass
        self.closed = True

    def get_data(self):
        input_stream = self.input_stream
        content_length = self.content_length
        content_type = self.content_type

        if input_stream is not None:
            start_pos = input_stream.tell()
            input_stream.seek(0, 2)
            end_pos = input_stream.tell()
            input_stream.seek(start_pos)
            content_length = end_pos - start_pos
        elif content_type == 'multipart/form-data':
            values = sdict()
            for d in [self.files, self.form]:
                for key, val in d.items():
                    values[key] = val
            input_stream, content_length, boundary = \
                stream_encode_multipart(values, charset=self.charset)
            content_type += '; boundary="%s"' % boundary
        elif content_type == 'application/x-www-form-urlencoded':
            values = url_encode(self.form, charset=self.charset)
            values = values.encode('ascii')
            content_length = len(values)
            input_stream = BytesIO(values)
        else:
            input_stream = BytesIO()

        result = {'headers': [(b'host', self.host.encode('utf-8'))]}
        if self.content_type:
            result['headers'].extend([
                (b'content-type', content_type.encode('utf-8')),
                (b'content-length', str(content_length or 0).encode('utf-8'))])
        if self.environ_base:
            result.update(self.environ_base)

        result.update({
            'type': 'http',
            'http_version': self.server_protocol,
            'server': (self.server_name, self.server_port),
            'client': (self.server_name, self.server_port),
            'method': self.method,
            'scheme': self.url_scheme,
            'root_path': self.script_root,
            'path': self.path,
            'query_string': self.query_string.encode('utf-8'),
            'emt.path': self.path,
            'emt.input': None
        })
        for key, value in self.headers.to_asgi_list():
            result['headers'].append(
                (key.encode('utf-8'), value.encode('utf-8')))
        if self.environ_overrides:
            result.update(self.environ_overrides)
        return result, input_stream.getvalue()



# -*- coding: utf-8 -*-
"""
    emmett.asgi.server
    ------------------

    Provides ASGI server wrapper over uvicorn

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import logging
import multiprocessing
import os
import signal
import socket
import ssl
import sys
import threading

from abc import ABCMeta, abstractmethod
from typing import Any, Optional, Tuple

from uvicorn.config import logger as uvlogger
from uvicorn.server import Server

from .._internal import locate_app
from ..extensions import Signals
from ..logger import LOG_LEVELS
from .helpers import Config
from .loops import loops
from .protocols import protocols_http, protocols_ws

multiprocessing.allow_connection_pickling()


class Runner(metaclass=ABCMeta):
    def __init__(
        self,
        app_target: Tuple[str, Optional[str]],
        workers: int,
        **kwargs: Any
    ):
        self.app_target = app_target
        self.workers = workers
        self.cfgdata = {**kwargs}

    @staticmethod
    def serve(app_target, cfgdata, sockets=None):
        cfg = {**cfgdata}
        cfg["loop"] = loops.get(cfg.pop("loop"))
        cfg["http"] = protocols_http.get(cfg.pop("proto_http"))
        cfg["ws"] = protocols_ws.get(cfg.pop("proto_ws"))

        app = locate_app(*app_target)
        app.send_signal(Signals.after_loop, loop=cfg["loop"])

        cfg["access_log"] = (
            cfg["access_log"] is not None and cfg["access_log"] or
            bool(app.debug)
        )
        cfg["log_level"] = LOG_LEVELS[cfg["log_level"]] if cfg["log_level"] else (
            logging.DEBUG if app.debug else logging.WARNING
        )
        cfg["forwarded_allow_ips"] = cfg.pop("proxy_trust_ips")

        Server(Config(app=app, **cfg)).run(sockets=sockets)

    @abstractmethod
    def run(self):
        ...


class SingleRunner(Runner):
    def run(self):
        self.serve(self.app_target, self.cfgdata)


class MultiRunner(Runner):
    SIGNALS = {
        signal.SIGINT,
        signal.SIGTERM
    }

    def bind_socket(self):
        family = socket.AF_INET
        if self.cfgdata["host"] and ":" in self.cfgdata["host"]:
            family = socket.AF_INET6

        sock = socket.socket(family=family)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            sock.bind((self.cfgdata["host"], self.cfgdata["port"]))
        except OSError as exc:
            uvlogger.error(exc)
            sys.exit(1)
        sock.set_inheritable(True)
        return sock

    def signal_handler(self, *args, **kwargs):
        self.exit_event.set()

    @staticmethod
    def _subprocess_target(target, data, stdin_fileno):
        if stdin_fileno is not None:
            sys.stdin = os.fdopen(stdin_fileno)
        target(**data)

    def _subprocess_spawn(self, target, **data):
        try:
            stdin_fileno = sys.stdin.fileno()
        except OSError:
            stdin_fileno = None

        return multiprocessing.get_context("spawn").Process(
            target=self._subprocess_target,
            kwargs={
                "target": target,
                "data": data,
                "stdin_fileno": stdin_fileno
            }
        )

    def startup(self):
        for sig in self.SIGNALS:
            signal.signal(sig, self.signal_handler)

        for _ in range(self.workers):
            proc = self._subprocess_spawn(
                target=self.serve,
                cfgdata=self.cfgdata,
                app_target=self.app_target,
                sockets=self.sockets
            )
            proc.start()
            self.processes.append(proc)

    def shutdown(self):
        for proc in self.processes:
            proc.join()

    def run(self):
        self.processes = []
        self.sockets = [self.bind_socket()]
        self.exit_event = threading.Event()

        self.startup()
        self.exit_event.wait()
        self.shutdown()


def run(
    app,
    host='127.0.0.1',
    port=8000,
    uds=None,
    fd=None,
    loop='auto',
    proto_http='auto',
    proto_ws='auto',
    log_level=None,
    access_log=None,
    proxy_headers=False,
    proxy_trust_ips=None,
    workers=1,
    limit_concurrency=None,
    # limit_max_requests=None,
    backlog=2048,
    timeout_keep_alive=0,
    # timeout_notify=30,
    ssl_certfile: Optional[str] = None,
    ssl_keyfile: Optional[str] = None,
    ssl_cert_reqs: int = ssl.CERT_NONE,
    ssl_ca_certs: Optional[str] = None
):
    if proxy_trust_ips is None:
        proxy_trust_ips = os.environ.get("PROXY_TRUST_IPS", "*")

    runner_cls = MultiRunner if workers > 1 else SingleRunner

    runner = runner_cls(
        app,
        workers,
        host=host,
        port=port,
        uds=uds,
        fd=fd,
        loop=loop,
        proto_http=proto_http,
        proto_ws=proto_ws,
        log_level=log_level,
        access_log=access_log,
        proxy_headers=proxy_headers,
        proxy_trust_ips=proxy_trust_ips,
        limit_concurrency=limit_concurrency,
        # limit_max_requests=None,
        backlog=backlog,
        timeout_keep_alive=timeout_keep_alive,
        # timeout_notify=30,
        ssl_certfile=ssl_certfile,
        ssl_keyfile=ssl_keyfile,
        ssl_cert_reqs=ssl_cert_reqs,
        ssl_ca_certs=ssl_ca_certs
    )
    runner.run()



# -*- coding: utf-8 -*-
"""
    emmett.parsers
    --------------

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from functools import partial
from typing import Any, Callable, Dict

try:
    import orjson
    _json_impl = orjson.loads
    _json_opts = {}
except ImportError:
    import rapidjson
    _json_impl = rapidjson.loads
    _json_opts = {
        "datetime_mode": rapidjson.DM_ISO8601 | rapidjson.DM_NAIVE_IS_UTC,
        "number_mode": rapidjson.NM_NATIVE
    }


class Parsers(object):
    _registry_: Dict[str, Callable[[str], Dict[str, Any]]] = {}

    @classmethod
    def register_for(cls, target):
        def wrap(f):
            cls._registry_[target] = f
            return f
        return wrap

    @classmethod
    def get_for(cls, target):
        return cls._registry_[target]


json = partial(
    _json_impl,
    **_json_opts
)

Parsers.register_for('json')(json)



# -*- coding: utf-8 -*-
"""
    emmett.orm.migrations.generation
    --------------------------------

    Provides generation utils for migrations.

    :copyright: 2014 Giovanni Barillari

    Based on the code of Alembic (https://bitbucket.org/zzzeek/alembic)
    :copyright: (c) 2009-2015 by Michael Bayer

    :license: BSD-3-Clause
"""

from collections import OrderedDict

from ...datastructures import OrderedSet
from .base import Column
from .helpers import Dispatcher, DEFAULT_VALUE, _feasible_as_dbms_default
from .operations import UpgradeOps, CreateTableOp, DropTableOp, \
    AddColumnOp, DropColumnOp, AlterColumnOp, CreateIndexOp, DropIndexOp


class MetaTable(object):
    def __init__(self, name, columns=[], indexes=[], **kw):
        self.name = name
        self.columns = OrderedDict()
        for column in columns:
            self.columns[column.name] = column
        self.indexes = {}
        self.kw = kw

    @property
    def fields(self):
        return list(self.columns)

    def __getitem__(self, name):
        return self.columns[name]

    def __setitem__(self, name, value):
        self.columns[name] = value

    def __delitem__(self, name):
        del self.columns[name]

    def __repr__(self):
        return "Table(%r, %s)" % (
            self.name,
            ", ".join(["%s" % column for column in self.columns.values()])
        )


class MetaIndex(object):
    def __init__(self, table_name, name, fields, expressions, unique, **kw):
        self.table_name = table_name
        self.name = name
        self.fields = fields
        self.expressions = expressions
        self.unique = unique
        self.kw = kw

    @property
    def where(self):
        return self.kw.get('where')

    def __repr__(self):
        opts = [('expressions', self.expressions), ('unique', self.unique)]
        for key, val in self.kw.items():
            opts.append((key, val))
        return "Index(%r, %r, %s)" % (
            self.name, self.fields,
            ", ".join(["%s=%r" % (opt[0], opt[1]) for opt in opts])
        )


class MetaData(object):
    def __init__(self):
        self.tables = {}

    def create_table(self, name, columns, **kw):
        self.tables[name] = MetaTable(name, columns)

    def drop_table(self, name):
        del self.tables[name]

    def add_column(self, table, column):
        self.tables[table][column.name] = column

    def drop_column(self, table, column):
        del self.tables[table][column]

    def change_column(self, table_name, column_name, changes):
        self.tables[table_name][column_name].update(**changes)

    def create_index(
        self, table_name, index_name, fields, expressions, unique, **kw
    ):
        self.tables[table_name].indexes[index_name] = MetaIndex(
            table_name, index_name, fields, expressions, unique, **kw
        )

    def drop_index(self, table_name, index_name):
        del self.tables[table_name].indexes[index_name]


class Comparator(object):
    def __init__(self, db, meta):
        self.db = db
        self.meta = meta

    def make_ops(self):
        self.ops = []
        self.tables()
        return self.ops

    def _build_metatable(self, dbtable):
        columns = []
        for field in list(dbtable):
            columns.append(Column.from_field(field))
        return MetaTable(
            dbtable._tablename,
            columns
        )

    def _build_metaindex(self, dbtable, index_name):
        model = dbtable._model_
        dbindex = model._indexes_[index_name]
        kw = {}
        with self.db._adapter.index_expander():
            if 'where' in dbindex:
                kw['where'] = str(dbindex['where'])
            rv = MetaIndex(
                model.tablename, index_name,
                [field for field in dbindex['fields']],
                [str(expr) for expr in dbindex['expressions']],
                dbindex['unique'], **kw
            )
        return rv

    def tables(self):
        db_table_names = OrderedSet([t._tablename for t in self.db])
        meta_table_names = OrderedSet(list(self.meta.tables))
        #: new tables
        for table_name in db_table_names.difference(meta_table_names):
            meta_table = self._build_metatable(self.db[table_name])
            self.ops.append(CreateTableOp.from_table(meta_table))
            self.indexes_and_uniques(self.db[table_name], meta_table)
        #: removed tables
        for table_name in meta_table_names.difference(db_table_names):
            #: remove table indexes too
            metatable = self.meta.tables[table_name]
            for idx_name, idx in metatable.indexes.items():
                self.ops.append(DropIndexOp.from_index(idx))
            #: remove table
            self.ops.append(
                DropTableOp.from_table(self.meta.tables[table_name]))
        #: existing tables
        for table_name in meta_table_names.intersection(db_table_names):
            self.columns(
                self.db[table_name], self.meta.tables[table_name])
            self.table(
                self.db[table_name], self.meta.tables[table_name])

    def table(self, dbtable, metatable):
        self.indexes_and_uniques(dbtable, metatable)
        self.foreign_keys(dbtable, metatable)

    def indexes_and_uniques(self, dbtable, metatable, ops_stack=None):
        ops = ops_stack if ops_stack is not None else self.ops
        db_index_names = OrderedSet(
            [idxname for idxname in dbtable._model_._indexes_.keys()])
        meta_index_names = OrderedSet(list(metatable.indexes))
        #: removed indexes
        for index_name in meta_index_names.difference(db_index_names):
            ops.append(DropIndexOp.from_index(metatable.indexes[index_name]))
        #: new indexs
        for index_name in db_index_names.difference(meta_index_names):
            ops.append(
                CreateIndexOp.from_index(
                    self._build_metaindex(dbtable, index_name)))
        #: existing indexes
        for index_name in meta_index_names.intersection(db_index_names):
            metaindex = metatable.indexes[index_name]
            dbindex = self._build_metaindex(dbtable, index_name)
            if any(
                getattr(metaindex, key) != getattr(dbindex, key)
                for key in ['fields', 'expressions', 'unique', 'kw']
            ):
                ops.append(DropIndexOp.from_index(metaindex))
                ops.append(CreateIndexOp.from_index(dbindex))
        # TODO: uniques

    def foreign_keys(self, dbtable, metatable, ops_stack=None):
        # TODO
        pass

    def columns(self, dbtable, metatable):
        db_column_names = OrderedSet([fname for fname in dbtable.fields])
        meta_column_names = OrderedSet(metatable.fields)
        #: new columns
        for column_name in db_column_names.difference(meta_column_names):
            self.ops.append(AddColumnOp.from_column_and_tablename(
                dbtable._tablename, Column.from_field(dbtable[column_name])))
        #: existing columns
        for column_name in meta_column_names.intersection(db_column_names):
            self.ops.append(AlterColumnOp(dbtable._tablename, column_name))
            self.column(
                dbtable[column_name], metatable.columns[column_name])
            if not self.ops[-1].has_changes():
                self.ops.pop()
        #: removed columns
        for column_name in meta_column_names.difference(db_column_names):
            self.ops.append(
                DropColumnOp.from_column_and_tablename(
                    dbtable._tablename, metatable.columns[column_name]))

    def column(self, dbcolumn, metacolumn):
        self.notnulls(dbcolumn, metacolumn)
        self.types(dbcolumn, metacolumn)
        self.lengths(dbcolumn, metacolumn)
        self.defaults(dbcolumn, metacolumn)

    def types(self, dbcolumn, metacolumn):
        self.ops[-1].existing_type = metacolumn.type
        if dbcolumn.type != metacolumn.type:
            self.ops[-1].modify_type = dbcolumn.type

    def lengths(self, dbcolumn, metacolumn):
        self.ops[-1].existing_length = metacolumn.length
        if any(
            field.type == "string" for field in [dbcolumn, metacolumn]
        ) and dbcolumn.length != metacolumn.length:
            self.ops[-1].modify_length = dbcolumn.length

    def notnulls(self, dbcolumn, metacolumn):
        self.ops[-1].existing_notnull = metacolumn.notnull
        if dbcolumn.notnull != metacolumn.notnull:
            self.ops[-1].modify_notnull = dbcolumn.notnull

    def defaults(self, dbcolumn, metacolumn):
        oldv, newv = metacolumn.default, dbcolumn.default
        self.ops[-1].existing_default = oldv
        if newv != oldv:
            if not all(callable(v) for v in [oldv, newv]):
                if _feasible_as_dbms_default(newv):
                    self.ops[-1].modify_default = newv

    @classmethod
    def compare(cls, db, meta):
        ops = cls(db, meta).make_ops()
        return UpgradeOps(ops)


class Generator(object):
    def __init__(self, db, scriptdir, head):
        self.db = db
        self.scriptdir = scriptdir
        self.head = head
        self.meta = MetaData()
        self._load_head_to_meta()

    def _load_head_to_meta(self):
        for revision in reversed(
            list(self.scriptdir.walk_revisions("base", self.head))
        ):
            migration = revision.migration_class(
                None, self.meta, is_meta=True)
            migration.up()

    def generate(self):
        return Comparator.compare(self.db, self.meta)

    @classmethod
    def generate_from(cls, dal, scriptdir, head):
        return cls(dal, scriptdir, head).generate()


class Renderer(object):
    def render_op(self, op):
        op_renderer = renderers.dispatch(op)
        return op_renderer(op)

    def render_opcontainer(self, op_container):
        rv = []
        if not op_container.ops:
            rv.append("pass")
        else:
            for op in op_container.ops:
                rv.append(self.render_op(op))
        return rv

    @classmethod
    def render_migration(cls, migration_op):
        r = cls()
        return r.render_opcontainer(migration_op.upgrade_ops), \
            r.render_opcontainer(migration_op.downgrade_ops)


renderers = Dispatcher()


@renderers.dispatch_for(CreateTableOp)
def _add_table(op):
    table = op.to_table()

    args = [
        col for col in [_render_column(col) for col in table.columns.values()]
        if col
    ]
    # + sorted([
    #     rcons for rcons in [
    #         _render_constraint(cons) for cons in table.constraints]
    #         if rcons is not None
    #     ]
    # )
    indent = " " * 12

    if len(args) > 255:
        args = '*[' + (',\n' + indent).join(args) + ']'
    else:
        args = (',\n' + indent).join(args)

    text = ("self.create_table(\n" + indent + "%(tablename)r,\n" + indent +
            "%(args)s") % {
        'tablename': op.table_name,
        'args': args
    }
    for k in sorted(op.kw):
        text += ",\n" + indent + "%s=%r" % (k.replace(" ", "_"), op.kw[k])
    text += ")"
    return text


@renderers.dispatch_for(DropTableOp)
def _drop_table(op):
    text = "self.drop_table(%(tname)r" % {
        "tname": op.table_name
    }
    text += ")"
    return text


def _render_column(column):
    opts = []

    if column.default is not None:
        opts.append(("default", column.default))

    # if not column.autoincrement:
    #     opts.append(("autoincrement", column.autoincrement))

    if column.notnull:
        opts.append(("notnull", column.notnull))

    if column.type in ("string", "password", "upload"):
        opts.append(("length", column.length))
    elif column.type.startswith('reference'):
        opts.append(("ondelete", column.ondelete))

    kw_str = ""
    if opts:
        kw_str = ", %s" % \
            ", ".join(["%s=%r" % (key, val) for key, val in opts])
    return "migrations.Column(%(name)r, %(type)r%(kw)s)" % {
        'name': column.name,
        'type': column.type,
        'kw': kw_str
    }


@renderers.dispatch_for(AddColumnOp)
def _add_column(op):
    return "self.add_column(%(tname)r, %(column)s)" % {
        "tname": op.table_name,
        "column": _render_column(op.column)
    }


@renderers.dispatch_for(DropColumnOp)
def _drop_column(op):
    return "self.drop_column(%(tname)r, %(cname)r)" % {
        "tname": op.table_name,
        "cname": op.column_name
    }


@renderers.dispatch_for(AlterColumnOp)
def _alter_column(op):
    indent = " " * 12
    text = "self.alter_column(%(tname)r, %(cname)r" % {
        'tname': op.table_name,
        'cname': op.column_name}

    if op.existing_type is not None:
        text += ",\n%sexisting_type=%r" % (indent, op.existing_type)
    if op.existing_length:
        text += ",\n%sexisting_length=%r" % (indent, op.existing_length)
    if op.modify_default is not DEFAULT_VALUE:
        text += ",\n%sdefault=%r" % (indent, op.modify_default)
    if op.modify_type is not None:
        text += ",\n%stype=%r" % (indent, op.modify_type)
    if op.modify_length is not None:
        text += ",\n%slength=%r" % (indent, op.modify_length)
    if op.modify_notnull is not None:
        text += ",\n%snotnull=%r" % (indent, op.modify_notnull)
    if op.modify_notnull is None and op.existing_notnull is not None:
        text += ",\n%sexisting_notnull=%r" % (indent, op.existing_notnull)
    if op.modify_default is DEFAULT_VALUE and op.existing_default:
        text += ",\n%sexisting_default=%s" % (indent, op.existing_default)

    text += ")"
    return text


@renderers.dispatch_for(CreateIndexOp)
def _add_index(op):
    kw_str = ""
    if op.kw:
        kw_str = ", %s" % ", ".join(
            ["%s=%r" % (key, val) for key, val in op.kw.items()])
    return "self.create_index(%(iname)r, %(tname)r, %(idata)s)" % {
        "tname": op.table_name,
        "iname": op.index_name,
        "idata": "%r, expressions=%r, unique=%s%s" % (
            op.fields, op.expressions, op.unique, kw_str)
    }


@renderers.dispatch_for(DropIndexOp)
def _drop_index(op):
    return "self.drop_index(%(iname)r, %(tname)r)" % {
        "tname": op.table_name,
        "iname": op.index_name
    }



# -*- coding: utf-8 -*-
"""
    emmett.templating.templater
    ---------------------------

    Provides the Emmett implementation for Renoir engine.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import os

from functools import reduce
from typing import Optional, Tuple

from renoir import Renoir

from .lexers import lexers


class Templater(Renoir):
    def __init__(self, **kwargs):
        kwargs['lexers'] = lexers
        super().__init__(**kwargs)
        self._namespaces = {}

    def _set_reload(self, value):
        self.cache.changes = value
        self.cache.load._configure()
        self.cache.prerender._configure()
        self.cache.parse._configure()

    def _set_encoding(self, value):
        self.encoding = value

    def _set_escape(self, value):
        self.escape = value
        self._configure()

    def _set_indent(self, value):
        self.indent = value
        self._configure()

    def register_namespace(self, namespace: str, path: Optional[str] = None):
        path = path or self.path
        self._namespaces[namespace] = path

    def _get_namespace_path_elements(self, file_name: str) -> Tuple[str, str]:
        if ":" in file_name:
            namespace, file_name = file_name.split(":")
            path = self._namespaces.get(namespace, self.path)
        else:
            path = self.path
        return path, file_name

    def _preload(self, file_name: str):
        path, file_name = self._get_namespace_path_elements(file_name)
        file_extension = os.path.splitext(file_name)[1]
        return reduce(
            lambda args, loader: loader(args[0], args[1]),
            self.loaders.get(file_extension, []),
            (path, file_name)
        )

    def _no_preload(self, file_name):
        return self._get_namespace_path_elements(file_name)



# -*- coding: utf-8 -*-
"""
    emmett.tools.decorators
    -----------------------

    Provides requires and service decorators.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from ..routing.router import Router


class Decorator(object):
    def build_pipe(self):
        pass

    def __call__(self, func):
        obj = Router.exposing()
        obj.pipeline.append(self.build_pipe())
        return func


class requires(Decorator):
    def __init__(self, condition=None, otherwise=None):
        if condition is None or otherwise is None:
            raise SyntaxError(
                'requires usage: @requires(condition, otherwise)'
            )
        if not callable(otherwise) and not isinstance(otherwise, str):
            raise SyntaxError(
                "requires 'otherwise' param must be string or callable"
            )
        self.condition = condition
        self.otherwise = otherwise

    def build_pipe(self):
        from ..pipeline import RequirePipe
        return RequirePipe(self.condition, self.otherwise)


class service(Decorator):
    def __init__(self, procedure):
        self.procedure = procedure

    @staticmethod
    def json(f):
        return service('json')(f)

    @staticmethod
    def xml(f):
        return service('xml')(f)

    def build_pipe(self):
        from .service import ServicePipe
        return ServicePipe(self.procedure)



# -*- coding: utf-8 -*-
"""
    emmett._shortcuts
    -----------------

    Some shortcuts

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import hashlib

hashlib_md5 = lambda s: hashlib.md5(bytes(s, 'utf8'))
hashlib_sha1 = lambda s: hashlib.sha1(bytes(s, 'utf8'))


def to_bytes(obj, charset='utf8', errors='strict'):
    if obj is None:
        return None
    if isinstance(obj, (bytes, bytearray, memoryview)):
        return bytes(obj)
    if isinstance(obj, str):
        return obj.encode(charset, errors)
    raise TypeError('Expected bytes')


def to_unicode(obj, charset='utf8', errors='strict'):
    if obj is None:
        return None
    if not isinstance(obj, bytes):
        return str(obj)
    return obj.decode(charset, errors)



# -*- coding: utf-8 -*-
"""
    tests.translator
    ----------------

    Test Emmett translator module
"""

import pytest

from emmett import App
from emmett.ctx import current
from emmett.locals import T


@pytest.fixture(scope='module')
def app():
    return App(__name__)


def _make_translation(language):
    return str(T('partly cloudy', lang=language))


def test_translation(app):
    current.language = 'en'
    assert _make_translation('it') == 'nuvolosità variabile'
    assert _make_translation('de') == 'teilweise bewölkt'
    assert _make_translation('ru') == 'переменная облачность'



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['emmett',
 'emmett.asgi',
 'emmett.asgi.loops',
 'emmett.asgi.protocols',
 'emmett.asgi.protocols.http',
 'emmett.asgi.protocols.ws',
 'emmett.language',
 'emmett.libs',
 'emmett.orm',
 'emmett.orm.engines',
 'emmett.orm.migrations',
 'emmett.routing',
 'emmett.templating',
 'emmett.testing',
 'emmett.tools',
 'emmett.tools.auth',
 'emmett.validators',
 'emmett.wrappers',
 'tests']

package_data = \
{'': ['*'],
 'emmett': ['assets/*', 'assets/debug/*'],
 'tests': ['languages/*', 'templates/*', 'templates/auth/*']}

install_requires = \
['click>=6.0',
 'h11>=0.10.0,<0.11.0',
 'h2>=3.2.0,<4.1.0',
 'pendulum>=2.1.2,<2.2.0',
 'pyDAL==17.3',
 'pyaes>=1.6.1,<1.7.0',
 'python-rapidjson>=1.0,<2.0',
 'pyyaml>=5.4,<6.0',
 'renoir>=1.3,<2.0',
 'severus>=1.1,<2.0',
 'uvicorn==0.14.0',
 'websockets>=9.1,<10.0']

extras_require = \
{':sys_platform != "win32"': ['httptools>=0.2.0,<0.3.0',
                              'uvloop>=0.15.3,<0.16.0'],
 'crypto': ['emmett-crypto>=0.1.0,<0.2.0'],
 'orjson': ['orjson>=3.5.1,<3.6.0']}

entry_points = \
{'console_scripts': ['emmett = emmett.cli:main']}

setup_kwargs = {
    'name': 'emmett',
    'version': '2.3.0',
    'description': 'The web framework for inventors',
    'long_description': '![Emmett](https://deneb.spaces.amira.io/emmett/artwork/logo-bwb-xb-xl.png)\n\nEmmett is a full-stack Python web framework designed with simplicity in mind.\n\nThe aim of Emmett is to be clearly understandable, easy to be learned and to be \nused, so you can focus completely on your product\'s features:\n\n```python\nfrom emmett import App, request, response\nfrom emmett.orm import Database, Model, Field\nfrom emmett.tools import service, requires\n\nclass Task(Model):\n    name = Field.string()\n    is_completed = Field.bool(default=False)\n\napp = App(__name__)\napp.config.db.uri = "postgres://user:password@localhost/foo"\ndb = Database(app)\ndb.define_models(Task)\napp.pipeline = [db.pipe]\n\ndef is_authenticated():\n    return request.headers["Api-Key"] == "foobar"\n    \ndef not_authorized():\n    response.status = 401\n    return {\'error\': \'not authorized\'}\n\n@app.route(methods=\'get\')\n@service.json\n@requires(is_authenticated, otherwise=not_authorized)\nasync def todo():\n    page = request.query_params.page or 1\n    tasks = Task.where(\n        lambda t: t.is_completed == False\n    ).select(paginate=(page, 20))\n    return {\'tasks\': tasks}\n```\n\n[![pip version](https://img.shields.io/pypi/v/emmett.svg?style=flat)](https://pypi.python.org/pypi/emmett)\n![Tests Status](https://github.com/emmett-framework/emmett/workflows/Tests/badge.svg)\n\n## Documentation\n\nThe documentation is available at [https://emmett.sh/docs](https://emmett.sh/docs).\nThe sources are available under the [docs folder](https://github.com/emmett-framework/emmett/tree/master/docs).\n\n## Examples\n\nThe *bloggy* example described in the [Tutorial](https://emmett.sh/docs/latest/tutorial) is available under the [examples folder](https://github.com/emmett-framework/emmett/tree/master/examples). \n\n## Status of the project\n\nEmmett is production ready and is compatible with Python 3.7 and above versions.\n\nEmmett follows a *semantic versioning* for its releases, with a `{major}.{minor}.{patch}` scheme for versions numbers, where:\n\n- *major* versions might introduce breaking changes\n- *minor* versions usually introduce new features and might introduce deprecations\n- *patch* versions only introduce bug fixes\n\nDeprecations are kept in place for at least 3 minor versions, and the drop is always communicated in the [upgrade guide](https://emmett.sh/docs/latest/upgrading).\n\n## How can I help?\n\nWe would be very glad if you contributed to the project in one or all of these ways:\n\n* Talking about Emmett with friends and on the web\n* Adding issues and features requests here on GitHub\n* Participating in discussions about new features and issues here on GitHub\n* Improving the documentation\n* Forking the project and writing beautiful code\n\n## License\n\nEmmmett is released under the BSD License.\n\nHowever, due to original license limitations, some components are included \nin Emmett under their original licenses. Please check the LICENSE file for \nmore details.\n',
    'author': 'Giovanni Barillari',
    'author_email': 'gi0baro@d4net.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://emmett.sh',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
